from mysql.connector import connection
from mysql.connector import Error

from time import sleep

from sqlIntuitive import sqlGeneration
from sqlIntuitive.ext.customDataTypes import AdaptionProvider, CustomDataType
from sqlIntuitive.dbSystems import BaseDbSystem
from sqlIntuitive.dbSystems.supportTracker import or_values, Features

class MySqlDbSystem(BaseDbSystem):
    placeholder = "%s"
    SUPPORTS = or_values([
        Features.SQL_CREATE_TABLE,
        Features.SQL_DROP_TABLE,
        Features.SQL_INSERT_INTO,
        Features.SQL_UPDATE,
        Features.SQL_DELETE_FROM,
        Features.SQL_SELECT_FROM,
        Features.SQL_COUNT_AVG_SUM,
        Features.SQL_PRIMARY_KEYS,
        Features.SQL_FOREIGN_KEYS,
        Features.SQL_FOREIGN_KEYS_CUSTOM_CONSTRAINT,
        Features.SQL_UNIQUE,
        Features.ADDON_CUSTOM_DATA_TYPE
        ])

    def __init__(self, host: str, database: str, username: str, password: str, adaptionProvider: AdaptionProvider = None, max_connect_retries: int = 5) -> None:
        super().__init__(database, adaptionProvider)

        self.host = host
        self.username = username
        self.password = password
        self.max_connect_retries = max_connect_retries

    def connect_to_db(self, retryNo: int = 0, timeout: int = 3) -> bool:
        if self.dbCon != None: return False

        try:
            self.dbCon = connection.MySQLConnection(
                host=self.host,
                database=self.database,
                username=self.username,
                password=self.password
            )
        except Error as e:
            sleep(timeout)

            if retryNo >= self.max_connect_retries or self.connect_to_db(retryNo+1) == False:
                return False

        return True

    def close_connection(self) -> None:
        if self.dbCon != None and self.dbCon.is_connected():
            self.dbCon.close()

    def create_cursor(self) -> None:
        if self.cursor != None or self.dbCon == None or self.dbCon.is_connected() == False:
            return

        self.cursor = self.dbCon.cursor()



from sqlIntuitive.exceptions.parameterExceptions.baseException import ParameterException

class InvalidType(ParameterException):
    def __init__(self, value, expectedType):
        super().__init__()

        self.value = value
        self.expectedType = expectedType

    def __str__(self):
        return f"'{self.value}' is not an instance of {self.expectedType}."

class DictionaryEmptyException(ParameterException):
    def __init__(self, text):
        super().__init__()

        self.text = text

    def __str__(self):
        return self.text

class InvalidTableNameException(ParameterException):
    def __init__(self):
        super().__init__()

    def __str__(self):
        return "Name for table contains invalid characters or is empty."

class InvalidDatabaseNameException(ParameterException):
    def __init__(self):
        super().__init__()

    def __str__(self):
        return "Name for database contains invalid characters or is empty."

class NotAClass(ParameterException):
    def __init__(self, expectedAsCls):
        super().__init__()

        self.cls = expectedAsCls

    def __str__(self):
        return f"{self.cls} is not a class."

class NotAFunction(ParameterException):
    def __init__(self, expectedAsFunc):
        super().__init__()

        self.func = expectedAsFunc

    def __str__(self):
        return f"{self.func} is not a function."

class NotAString(ParameterException):
    def __init__(self, expectedAsString):
        super().__init__()

        self.string = expectedAsString

    def __str__(self):
        return f"{self.string} is not a string."

class NotAMatchingClass(ParameterException):
    def __init__(self, givenType, expectedType):
        super().__init__()

        self.givenType = givenType
        self.expectedType = expectedType

    def __str__(self):
        return f"{self.givenType} does not match {self.expectedType}."



from sqlIntuitive.exceptions.parameterExceptions.subexceptions.invalidConstraintExceptions.baseException import InvalidConstraint
from sqlIntuitive.exceptions.parameterExceptions.subexceptions.invalidConstraintExceptions.subexceptions import *



import sqlite3

from sqlIntuitive import sqlGeneration
from sqlIntuitive.ext.customDataTypes import AdaptionProvider, CustomDataType
from sqlIntuitive.dbSystems import BaseDbSystem
from sqlIntuitive.dbSystems.supportTracker import or_values, Features

class SqliteDbSystem(BaseDbSystem):
    placeholder="?"
    SUPPORTS = or_values([
        Features.SQL_CREATE_TABLE,
        Features.SQL_DROP_TABLE,
        Features.SQL_INSERT_INTO,
        Features.SQL_UPDATE,
        Features.SQL_DELETE_FROM,
        Features.SQL_SELECT_FROM,
        Features.SQL_COUNT_AVG_SUM,
        Features.SQL_PRIMARY_KEYS,
        Features.SQL_FOREIGN_KEYS,
        Features.SQL_FOREIGN_KEYS_CUSTOM_CONSTRAINT,
        Features.SQL_UNIQUE,
        Features.ADDON_CUSTOM_DATA_TYPE
        ])

    def __init__(self, database: str, timeout: int = 5, adaptionProvider: AdaptionProvider = None):
        super().__init__(database, adaptionProvider)

        self.timeout = timeout
        self.open = False

    def addDataType(self, dataType: CustomDataType) -> None:
        self.adaptProvider.addDataType(dataType)

    def addDataType_raw(self, name: str, cls, clsToStringFunc, stringToClsFunc) -> None:
        self.adaptProvider.addDataType_raw(name, cls, clsToStringFunc, stringToClsFunc)

    def connect_to_db(self) -> None:
        if self.open == False:
            self.dbCon = sqlite3.connect(self.database, self.timeout)

            self.open = True

    def close_connection(self) -> None:
        if self.open != True:
            return

        if self.cursor != None:
            self.cursor.close()

        if self.dbCon != None:
            self.dbCon.close()

        self.open = False

    def create_cursor(self) -> None:
        if self.cursor != None or self.dbCon == None or self.open == False:
            return

        self.cursor = self.dbCon.cursor()



from sqlIntuitive.dbSystems import supportTracker

import unittest
import os, json

defaultFile = 'tests/testsToRun.json'
altFile = 'tests/testsToRunDefault.json'

with open((defaultFile if os.path.exists(defaultFile) else altFile), 'r') as file:
    runTestSupportTracker = json.load(file)['supportTracker']

@unittest.skipIf(runTestSupportTracker == False, 'Skipped TestSupportTracker via config')
class TestSupportTracker(unittest.TestCase):
    def connect_by_OR(self, numbers):
        sum = 0
        for number in numbers:
            sum = sum | number

        return sum

    def test_A_Features(self):
        featuresClassDict = supportTracker.Features.__dict__

        sqlFeatures = [featuresClassDict[i].value for i in featuresClassDict if i.startswith("SQL_")]
        addonFeatures = [featuresClassDict[i].value for i in featuresClassDict if i.startswith("ADDON_")]
        featureGroups = [featuresClassDict[i].value for i in featuresClassDict if i.startswith("GROUP_")]

        self.assertEqual(supportTracker.Features.GROUP_SQL.value, self.connect_by_OR(sqlFeatures))
        self.assertEqual(supportTracker.Features.GROUP_ADDON.value, self.connect_by_OR(addonFeatures))

        self.assertEqual(supportTracker.Features.ALL.value, self.connect_by_OR(featureGroups))



from sqlIntuitive import dbSystems
from sqlIntuitive.ext import customDataTypes
from sqlIntuitive.conditionEnums import ComparisonTypes, CombinationTypes

from sqlIntuitive import exceptions

import unittest

import json, os

defaultFile = 'tests/testsToRun.json'
altFile = 'tests/testsToRunDefault.json'

with open((defaultFile if os.path.exists(defaultFile) else altFile), 'r') as file:
    runTestMysqlSystem = json.load(file)['mysqlSystem']

@unittest.skipIf(runTestMysqlSystem == False, 'Skipped TestMysqlSystem via config')
class TestMysqlSystem(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        with open('tests/mysql_testserver_login.json', 'r') as file:
            cls.mysql_login = json.load(file)

        mydb = dbSystems.MySqlDbSystem(
            host=cls.mysql_login["host"],
            database=cls.mysql_login["database"],
            username=cls.mysql_login["username"],
            password=cls.mysql_login["password"],
        )

        mydb.connect_to_db()

        if mydb.dbCon.is_connected():
            cursor = mydb.dbCon.cursor()

            cursor.execute("CREATE TABLE IF NOT EXISTS TableB (col1 varchar(200), col2 int PRIMARY KEY, col3 tinyint, col4 int);")
        else:
            raise Exception()

        if mydb.dbCon != None and mydb.dbCon.is_connected():
            mydb.dbCon.close()

    def setUp(self):
        self.mydb = dbSystems.MySqlDbSystem(
            host=self.mysql_login["host"],
            database=self.mysql_login["database"],
            username=self.mysql_login["username"],
            password=self.mysql_login["password"],
        )

    def tearDown(self):
        if self.mydb.dbCon != None and self.mydb.dbCon.is_connected():
            self.mydb.dbCon.close()

        self.setUp()
        self.clearTableB()
        self.mydb.dbCon.close()

    def clearTableB(self):
        self.mydb.connect_to_db()
        cursor = self.mydb.dbCon.cursor()

        cursor.execute("DELETE FROM TableB;")
        self.mydb.dbCon.commit()

    def test_A_connect_to_db(self):
        self.assertTrue(self.mydb.connect_to_db())
        self.assertIsNotNone(self.mydb.dbCon)
        self.assertTrue(self.mydb.dbCon.is_connected())

        self.mydb.close_connection()

        self.assertFalse(self.mydb.dbCon.is_connected())

    def test_B_close_connection(self):
        self.assertIsNone(self.mydb.dbCon)

        self.mydb.close_connection()

        self.assertTrue(self.mydb.connect_to_db())

        self.mydb.close_connection()

        self.mydb.close_connection() # No close happens, but no error should appear.

    def test_C_get_cursor(self):
        self.assertTrue(self.mydb.connect_to_db())

        self.assertIsNone(self.mydb.cursor)

        self.mydb.create_cursor()

        self.assertIsNotNone(self.mydb.cursor)

        self.mydb.close_connection()

    def test_D_create_table(self):
        self.assertTrue(self.mydb.connect_to_db())
        self.mydb.create_cursor()
        self.assertIsNotNone(self.mydb.cursor)

        self.mydb.create_table("TestA", {"name":"varchar(50)","id": "int", "id2": "int", "col3": 'int', "col4": "varchar(50)", "col5": "int"}, primaryKeys=['id', 'id2'], foreignKeys={'col3': 'TableB(col2)'}, namedForeignKeys={'col5': {'name':'myTestFK', 'reference': "TableB(col2)"}}, uniqueColumns=['col4'])

    def test_E_drop_table(self):
        self.assertTrue(self.mydb.connect_to_db())
        self.mydb.create_cursor()
        self.assertIsNotNone(self.mydb.cursor)

        self.mydb.drop_table("TestA")

    def test_F_insert_into(self):
        self.assertTrue(self.mydb.connect_to_db())
        self.mydb.create_cursor()
        self.assertIsNotNone(self.mydb.cursor)

        self.mydb.insert_into("TableB", {"col1": 'Test', "col2": 42, "col3": True})
        self.mydb.insert_into("TableB", {"col1": 'Test', "col2": 84, "col3": True})

    def test_G_update(self):
        self.assertTrue(self.mydb.connect_to_db())
        self.mydb.create_cursor()
        self.assertIsNotNone(self.mydb.cursor)

        self.mydb.update("TableB", {"col3": False, "col2": 43}, {"col2": 42})
        self.mydb.update("TableB", {"col3": False, "col2": 44}, {"col2": {'value': 43, 'comparison': ComparisonTypes.LESS_THAN_OR_EQUAL_TO}})

    def test_H_select_from(self):
        self.assertTrue(self.mydb.connect_to_db())
        self.mydb.create_cursor()
        self.assertIsNotNone(self.mydb.cursor)

        self.mydb.select_from("TableB")
        self.mydb.cursor.execute("INSERT INTO TableB (col1,col2) VALUES ('A',10),('B',20),('A',30);")

        res = self.mydb.select_from("TableB", ['col1'], conditions={'col2': {'value':50, 'comparison': ComparisonTypes.LESS_THAN_OR_EQUAL_TO}})
        self.assertEqual(len(res), 3)

        res = self.mydb.select_from("TableB", ['col1'], conditions={'col2': {'value':50, 'comparison': ComparisonTypes.LESS_THAN_OR_EQUAL_TO}}, distinct=True)
        self.assertEqual(len(res), 2)

    def test_I_delete(self):
        self.assertTrue(self.mydb.connect_to_db())
        self.mydb.create_cursor()
        self.assertIsNotNone(self.mydb.cursor)

        self.mydb.delete_from("TableB", {"col2": 84})
        self.mydb.delete_from("TableB", {"col2": {'value': 80, 'comparison': ComparisonTypes.EQUAL_TO}})
        self.mydb.delete_from("TableB")

    def test_J_adaptionProvider(self):
        self.assertTrue(isinstance(self.mydb.adaptProvider, customDataTypes.AdaptionProvider))

        adaptProvider = customDataTypes.AdaptionProvider()
        mydb2 = dbSystems.MySqlDbSystem(
            host=self.mysql_login["host"],
            database=self.mysql_login["database"],
            username=self.mysql_login["username"],
            password=self.mysql_login["password"],
            adaptionProvider=adaptProvider
        )

        self.assertEqual(mydb2.adaptProvider, adaptProvider)

    def test_K_str_encoded(self):
        self.assertTrue(isinstance(self.mydb.adaptProvider, customDataTypes.AdaptionProvider))

        self.mydb.connect_to_db()
        self.mydb.create_cursor()

        self.mydb.insert_into("TableB", {'col1': "CUSTOM;", 'col2': 6473573})

        res = self.mydb.select_from("TableB", columns=['col1'], conditions={'col2': 6473573})
        self.assertEqual(res, [("CUSTOM;",)])

    def test_L_customDataType(self):
        self.assertTrue(isinstance(self.mydb.adaptProvider, customDataTypes.AdaptionProvider))

        class TestClass():
            def __init__(self, num):
                self.num = num

        def testClassToString(instance):
            return str(instance.num)

        def stringToTestClass(string):
            return TestClass(int(string))

        dataType = customDataTypes.CustomDataType("TESTCLS", TestClass, testClassToString, stringToTestClass)

        self.mydb.addDataType(dataType)

        self.assertTrue('TESTCLS' in self.mydb.adaptProvider.types)

        self.mydb.connect_to_db()
        self.mydb.create_cursor()

        self.mydb.insert_into("TableB", {'col1': TestClass(123), 'col2': 567})
        res = self.mydb.select_from("TableB", conditions={'col2': 567})

        self.assertTrue(isinstance(res[0][0], TestClass))

    def test_M_select_count(self):
        self.mydb.connect_to_db()
        self.mydb.create_cursor()

        self.mydb.insert_into("TableB", {'col1': "ABC", 'col2': 1, 'col4': 123})
        self.mydb.insert_into("TableB", {'col1': "DEF", 'col2': 2, 'col4': 456})
        self.mydb.insert_into("TableB", {'col1': "ABC", 'col2': 3, 'col4': 789})

        res = self.mydb.select_count("TableB", column="col4")
        self.assertEqual(res, 3)

        res = self.mydb.select_count("TableB", column="col4", conditions={"col1": "ABC"})
        self.assertEqual(res, 2)

        res = self.mydb.select_count("TableB", column="col4", conditions={"col4": {"value": 123, "comparison": ComparisonTypes.GREATER_THAN}})
        self.assertEqual(res, 2)


        res = self.mydb.select_count("TableB", column="col1")
        self.assertEqual(res, 3)

        res = self.mydb.select_count("TableB", column="col1", distinct=True)
        self.assertEqual(res, 2)

    def test_N_select_avg(self):
        self.mydb.connect_to_db()
        self.mydb.create_cursor()

        self.mydb.insert_into("TableB", {'col1': "ABC", 'col2': 1, 'col4': 123})
        self.mydb.insert_into("TableB", {'col1': "DEF", 'col2': 2, 'col4': 456})
        self.mydb.insert_into("TableB", {'col1': "ABC", 'col2': 3, 'col4': 789})

        res = self.mydb.select_avg("TableB", column="col4")
        self.assertEqual(res, 456)

        res = self.mydb.select_avg("TableB", column="col4", conditions={"col1": "ABC"})
        self.assertEqual(res, (123+789)/2)

        res = self.mydb.select_avg("TableB", column="col4", conditions={"col4": {"value": 123, "comparison": ComparisonTypes.GREATER_THAN}})
        self.assertEqual(res, (456+789)/2)


        self.mydb.insert_into("TableB", {'col1': "GHI", 'col2': 4, 'col4': 123})

        res = self.mydb.select_avg("TableB", column="col4")
        self.assertEqual(res, (123+456+789+123)/4)

        res = self.mydb.select_avg("TableB", column="col4", distinct=True)
        self.assertEqual(res, (123+456+789)/3)

    def test_O_select_sum(self):
        self.mydb.connect_to_db()
        self.mydb.create_cursor()

        self.mydb.insert_into("TableB", {'col1': "ABC", 'col2': 1, 'col4': 123})
        self.mydb.insert_into("TableB", {'col1': "DEF", 'col2': 2, 'col4': 456})
        self.mydb.insert_into("TableB", {'col1': "ABC", 'col2': 3, 'col4': 789})

        res = self.mydb.select_sum("TableB", column="col4")
        self.assertEqual(res, 123+456+789)

        res = self.mydb.select_sum("TableB", column="col4", conditions={"col1": "ABC"})
        self.assertEqual(res, 123+789)

        res = self.mydb.select_sum("TableB", column="col4", conditions={"col4": {"value": 123, "comparison": ComparisonTypes.GREATER_THAN}})
        self.assertEqual(res, 456+789)


        self.mydb.insert_into("TableB", {'col1': "GHI", 'col2': 4, 'col4': 123})

        res = self.mydb.select_sum("TableB", column="col4")
        self.assertEqual(res, 123+456+789+123)

        res = self.mydb.select_sum("TableB", column="col4", distinct=True)
        self.assertEqual(res, 123+456+789)

    def test_P_cursorIsNone(self):
        with self.assertRaises(exceptions.CursorIsNone):
            self.mydb.create_table("TableB", {"col1": "string"})

        with self.assertRaises(exceptions.CursorIsNone):
            self.mydb.drop_table("TableB")

        with self.assertRaises(exceptions.CursorIsNone):
            self.mydb.insert_into("TableB", {"col1": "abc"})

        with self.assertRaises(exceptions.CursorIsNone):
            self.mydb.update("TableB", {"col1": "abc"})

        with self.assertRaises(exceptions.CursorIsNone):
            self.mydb.delete_from("TableB")

        with self.assertRaises(exceptions.CursorIsNone):
            self.mydb.select_from("TableB")

        with self.assertRaises(exceptions.CursorIsNone):
            self.mydb.select_count("TableB", "col1")

        with self.assertRaises(exceptions.CursorIsNone):
            self.mydb.select_avg("TableB", "col1")

        with self.assertRaises(exceptions.CursorIsNone):
            self.mydb.select_sum("TableB", "col1")



from enum import Enum
from sqlIntuitive.exceptions import NotSupported

EXCEPTION_IF_NOT_SUPPORTED = True

def reuseLastReturn(func):
    lastReturn = None

    def replaceFunc(*args, **kwargs):
        nonlocal lastReturn

        newReturn = func(lastReturn, *args, **kwargs)
        lastReturn = newReturn

        return newReturn

    return replaceFunc

@reuseLastReturn
def nextPowerOf2(lastReturn: int):
    if lastReturn == None: return 1

    return 2 * lastReturn

class Features(Enum):
    # SQL language features
    SQL_CREATE_TABLE = nextPowerOf2()
    SQL_DROP_TABLE = nextPowerOf2()
    SQL_INSERT_INTO = nextPowerOf2()
    SQL_UPDATE = nextPowerOf2()
    SQL_DELETE_FROM = nextPowerOf2()
    SQL_SELECT_FROM = nextPowerOf2()
    SQL_COUNT_AVG_SUM = nextPowerOf2()

    SQL_PRIMARY_KEYS = nextPowerOf2()
    SQL_FOREIGN_KEYS = nextPowerOf2()
    SQL_FOREIGN_KEYS_CUSTOM_CONSTRAINT = nextPowerOf2()
    SQL_UNIQUE = nextPowerOf2()

    SQL_STORED_PROCEDURES = nextPowerOf2()

    # Addon features
    ADDON_CUSTOM_DATA_TYPE = nextPowerOf2()

    # GROUPS
    GROUP_SQL = SQL_CREATE_TABLE | SQL_DROP_TABLE | SQL_INSERT_INTO | SQL_UPDATE | SQL_DELETE_FROM | SQL_SELECT_FROM | SQL_PRIMARY_KEYS | SQL_FOREIGN_KEYS | SQL_FOREIGN_KEYS_CUSTOM_CONSTRAINT | SQL_UNIQUE | SQL_COUNT_AVG_SUM | SQL_STORED_PROCEDURES
    GROUP_ADDON = ADDON_CUSTOM_DATA_TYPE

    # all
    ALL = GROUP_SQL | GROUP_ADDON

def or_values(features: list):
    value = 0
    for feature in features:
        value = value | feature.value

    return value

def isSupported(feature: Features, supported: int) -> bool:
    if (supported & feature.value) == feature.value:
        return True
    return False

def ifSupported(feature: Features):

    def decorator(func):
        def replaceFunc(self, *args, **kwargs):
            supported = False
            name = self.__class__.__name__
            supports = self.__class__.SUPPORTS

            if isSupported(feature, supports):
                supported = True

            if supported:
                return func(self, *args, **kwargs)
            else:
                if EXCEPTION_IF_NOT_SUPPORTED:
                    raise NotSupported(feature.name, name)
                else:
                    return None


        return replaceFunc

    return decorator



from sqlIntuitive.dbSystems.baseDbSystem import BaseDbSystem
from sqlIntuitive.dbSystems.mysqlSystem import MySqlDbSystem
from sqlIntuitive.dbSystems.sqliteSystem import SqliteDbSystem
from sqlIntuitive.dbSystems import supportTracker



from sqlIntuitive.exceptions.baseException import SqlIntuitiveException

class ConditionException(SqlIntuitiveException):
    def __str__(self):
        return 'An overall \'Condition exception\' was raised processing a condition.'



from sqlIntuitive import exceptions
from sqlIntuitive.sqlGeneration import mysql

import unittest

import json, os

defaultFile = 'tests/testsToRun.json'
altFile = 'tests/testsToRunDefault.json'

with open((defaultFile if os.path.exists(defaultFile) else altFile), 'r') as file:
    runTestSqlGenerationMySQL = json.load(file)['sqlGeneration_mysql']

@unittest.skipIf(runTestSqlGenerationMySQL == False, 'Skipped TestSqlGenerationMySQL via config')
class TestSqlGenerationMySQL(unittest.TestCase):
    def test_A_a_gen_create_stored_procedure(self):
        self.assertEqual(mysql.gen_create_stored_procedure("TestProc", "SELECT * FROM TableA;", "$$"), "DELIMITER $$\nCREATE PROCEDURE TestProc ()\nBEGIN\nSELECT * FROM TableA;\nEND$$\nDELIMITER ;")
        self.assertEqual(mysql.gen_create_stored_procedure("TestProc", "SELECT * FROM TableA;", "$$", {'param1': 'varchar(10)'}), "DELIMITER $$\nCREATE PROCEDURE TestProc (IN param1 varchar(10))\nBEGIN\nSELECT * FROM TableA;\nEND$$\nDELIMITER ;")
        self.assertEqual(mysql.gen_create_stored_procedure("TestProc", "SELECT * FROM TableA;", "$$", {'param1': 'varchar(10)', 'param2': 'int'}), "DELIMITER $$\nCREATE PROCEDURE TestProc (IN param1 varchar(10), IN param2 int)\nBEGIN\nSELECT * FROM TableA;\nEND$$\nDELIMITER ;")

    def test_A_b_gen_create_stored_procedure(self):
        with self.assertRaises(exceptions.InvalidType):
            mysql.gen_create_stored_procedure(123, "SELECT * FROM TableA;", "$$")

        with self.assertRaises(exceptions.InvalidType):
            mysql.gen_create_stored_procedure("Test", 456, "$$")

        with self.assertRaises(exceptions.InvalidType):
            mysql.gen_create_stored_procedure("Test", "SELECT * FROM TableA;", 123)

        with self.assertRaises(exceptions.InvalidType):
            mysql.gen_create_stored_procedure("Test", "SELECT * FROM TableA;", "$$", ['abc','def'])

        with self.assertRaises(exceptions.InvalidType):
            mysql.gen_create_stored_procedure("Test", "SELECT * FROM TableA;", "$$", {123:'456'})

        with self.assertRaises(exceptions.InvalidType):
            mysql.gen_create_stored_procedure("Test", "SELECT * FROM TableA;", "$$", {'123':456})

    def test_B_a_gen_exec_procedure(self):
        self.assertEqual(mysql.gen_exec_procedure("TestProc"), ("CALL TestProc ();", []))
        self.assertEqual(mysql.gen_exec_procedure("TestProc", ['2022-01-04']), ("CALL TestProc (?);", ['2022-01-04']))
        self.assertEqual(mysql.gen_exec_procedure("TestProc", ['2022-01-04',123,'HelloWorld!']), ("CALL TestProc (?, ?, ?);", ['2022-01-04',123,'HelloWorld!']))

    def test_B_b_gen_exec_procedure(self):
        with self.assertRaises(exceptions.InvalidType):
            mysql.gen_exec_procedure(123)

        with self.assertRaises(exceptions.InvalidType):
            mysql.gen_exec_procedure("Test", ('abc','def'))



from sqlIntuitive import exceptions

def gen_create_stored_procedure(procedureName: str, sql_statement: str, delimiter: str, parameters: dict = {}) -> str:
    if not isinstance(procedureName, str):
        raise exceptions.InvalidType(procedureName, str)

    if not isinstance(sql_statement, str):
        raise exceptions.InvalidType(sql_statement, str)

    if not isinstance(delimiter, str):
        raise exceptions.InvalidType(delimiter, str)

    if not isinstance(parameters, dict):
        raise exceptions.InvalidType(parameters, dict)
    else:
        for parameter in parameters:
            if not isinstance(parameter, str) or not isinstance(parameters[parameter], str):
                raise exceptions.InvalidType("Dictionary key and/or value", str)

    text = f"DELIMITER {delimiter}\nCREATE PROCEDURE {procedureName} ("

    if len(parameters) > 0:
        text += ', '.join([f'IN {key} {parameters[key]}' for key in parameters])

    text += f")\nBEGIN\n{sql_statement}\nEND{delimiter}\nDELIMITER ;"

    return text

def gen_exec_procedure(procedureName: str, parameters: list = [], placeholder: str = '?') -> tuple:
    if not isinstance(procedureName, str):
        raise exceptions.InvalidType(procedureName, str)

    if not isinstance(parameters, list):
        raise exceptions.InvalidType(parameters, list)

    text = f"CALL {procedureName} ("

    values_ordered = []
    if len(parameters) > 0:
        paramList = ['?' for _ in range(len(parameters))]
        for parameter in parameters:
            values_ordered.append(parameter)

        text += ', '.join(paramList)

    text += ");"

    return text, values_ordered



from sqlIntuitive import exceptions

import unittest

import json, os

defaultFile = 'tests/testsToRun.json'
altFile = 'tests/testsToRunDefault.json'

with open((defaultFile if os.path.exists(defaultFile) else altFile), 'r') as file:
    runTestExceptions = json.load(file)['exceptions']

@unittest.skipIf(runTestExceptions == False, 'Skipped TestExceptions via config')
class TestExceptions(unittest.TestCase):
    def test_A_staticExceptionMessages(self):
        with self.assertRaisesRegex(exceptions.SqlIntuitiveException, 'An overall exception was raised inside the modul \'sqlIntuitive\'.'):
            raise exceptions.SqlIntuitiveException()

        with self.assertRaisesRegex(exceptions.AdaptionProviderException, 'An overall exception was raised inside the class \'sqlIntuitive.ext.customDataTypes.AdaptionProvider\'.'):
            raise exceptions.AdaptionProviderException()

        with self.assertRaisesRegex(exceptions.ConditionException, 'An overall \'Condition exception\' was raised processing a condition.'):
            raise exceptions.ConditionException()

        with self.assertRaisesRegex(exceptions.ParameterException, 'An overall \'Parameter exception\' was raised.'):
            raise exceptions.ParameterException()

        with self.assertRaisesRegex(exceptions.InvalidConstraint, 'An overall \'Invalid constraint\' parameter exception was raised.'):
            raise exceptions.InvalidConstraint()

        with self.assertRaisesRegex(exceptions.InvalidTableNameException, 'Name for table contains invalid characters or is empty.'):
            raise exceptions.InvalidTableNameException()

        with self.assertRaisesRegex(exceptions.InvalidDatabaseNameException, 'Name for database contains invalid characters or is empty.'):
            raise exceptions.InvalidDatabaseNameException()

    def test_B_adaptionProviderExceptions(self):
        with self.assertRaisesRegex(exceptions.DuplicationError, '.* (.*) are already registered'):
            raise exceptions.DuplicationError("TEST", 123)

        with self.assertRaisesRegex(exceptions.DeletingTypeNotAllowed, r'It is not allowed to remove .*\.'):
            raise exceptions.DeletingTypeNotAllowed("TEST")

    def test_C_conditionExceptions(self):
        with self.assertRaisesRegex(exceptions.NotAComparisonType, r'.* is not a comparison type\.'):
            raise exceptions.NotAComparisonType(type(123))

        with self.assertRaisesRegex(exceptions.NotACombinationType, r'.* is not a combination type\.'):
            raise exceptions.NotACombinationType(type(123))

        with self.assertRaisesRegex(exceptions.NoValue, r'No value for .*\.'):
            raise exceptions.NoValue("abc")

    def test_D_parameterExceptions(self):
        with self.assertRaisesRegex(exceptions.InvalidType, r'\'.*\' is not an instance of .*\.'):
            raise exceptions.InvalidType("abc", int)

        with self.assertRaisesRegex(exceptions.DictionaryEmptyException, 'Text'):
            raise exceptions.DictionaryEmptyException("Text")

    def test_E_invalidConstraints(self):
        with self.assertRaisesRegex(exceptions.InvalidPrimaryKeyColumn, r'.* not in columns\.'):
            raise exceptions.InvalidPrimaryKeyColumn("abc")

        with self.assertRaisesRegex(exceptions.InvalidForeignKeyColumn, r'.* not in columns\.'):
            raise exceptions.InvalidForeignKeyColumn("abc")

        with self.assertRaisesRegex(exceptions.InvalidUniqueColumn, r'.* not in columns\.'):
            raise exceptions.InvalidUniqueColumn("abc")

        with self.assertRaisesRegex(exceptions.InvalidNamedForeignKeyDictionary, r'Values for column .* are not defined\.'):
            raise exceptions.InvalidNamedForeignKeyDictionary("abc")

        with self.assertRaisesRegex(exceptions.NotAClass, r'.* is not a class\.'):
            raise exceptions.NotAClass("abc")

        with self.assertRaisesRegex(exceptions.NotAFunction, r'.* is not a function\.'):
            raise exceptions.NotAFunction("abc")

        with self.assertRaisesRegex(exceptions.NotAString, r'.* is not a string\.'):
            raise exceptions.NotAString("abc")

        with self.assertRaisesRegex(exceptions.NotAMatchingClass, r'.* does not match .*\.'):
            raise exceptions.NotAMatchingClass(int, str)



from sqlIntuitive.exceptions.baseException import SqlIntuitiveException

class ParameterException(SqlIntuitiveException):
    def __str__(self):
        return "An overall \'Parameter exception\' was raised."



from sqlIntuitive import exceptions
from sqlIntuitive.conditionEnums import ComparisonTypes, CombinationTypes
from enum import Enum, auto

import string
import re

INVALID_CHARS = ['!', '"', '#', r'\$', '%', '&', "'", r'\(', r'\)', r'\*', r'\+', ',', '-', '/', ':', ';', '<', '=', '>', r'\?', '@', r'\[', r'\\', r'\]', r'\^', '_', '`', r'\{', r'\|', r'\}', '~', ' ', '\n', '\t']

class Count_avg_sum_modes(Enum):
    COUNT = auto()
    AVG = auto()
    SUM = auto()

def check_validName(text: str) -> bool:
    for char in INVALID_CHARS:
        if re.match(f'.*{char}.*', text, re.I):
            return False

    return True

def gen_conditions(conditions: dict = {}, combinations: list = [], *, defaultCombination: CombinationTypes = CombinationTypes.AND, defaultComparison: ComparisonTypes = ComparisonTypes.EQUAL_TO, placeholder: str = "?") -> tuple:
    if type(conditions) != dict:
        raise exceptions.InvalidType(conditions, dict)

    if type(combinations) != list:
        raise exceptions.InvalidType(combinations, list)

    if type(defaultCombination) != CombinationTypes:
        try:
            defaultCombination = CombinationTypes(defaultCombination)
        except ValueError:
            raise exceptions.NotACombinationType(type(defaultCombination))

    if type(defaultComparison) != ComparisonTypes:
        try:
            defaultComparison = ComparisonTypes(defaultComparison)
        except ValueError:
            raise exceptions.NotAComparisonType(type(defaultComparison))

    combinations = combinations.copy()
    combinations_new = []

    for combination in combinations:
        if type(combination) != CombinationTypes:
            try:
                combination = CombinationTypes(combination)
            except ValueError:
                raise exceptions.NotACombinationType(f'{type(combination)} is not a combination type')
        combinations_new.append(combination)

    combinations = combinations_new
    del combinations_new

    conditionTexts = []
    values_ordered = []

    text = ""

    for column in conditions.keys():
        if type(conditions[column]) == dict:
            if 'value' not in conditions[column].keys():
                raise exceptions.NoValue(column)

            if 'comparison' in conditions[column].keys():
                comparison = conditions[column]['comparison']
                if type(comparison) != ComparisonTypes:
                    try:
                        comparison = ComparisonTypes(comparison)
                    except ValueError:
                        raise exceptions.NotAComparisonType(f'{type(conditions[column]["comparison"])} is not a comparison type.')

                conditionTexts.append(f'{column}{comparison.value}{placeholder}')
                values_ordered.append(conditions[column]["value"])

            else:
                conditionTexts.append(f'{column}{defaultComparison.value}{placeholder}')
                values_ordered.append(conditions[column]["value"])
        else:
            conditionTexts.append(f'{column}{defaultComparison.value}{placeholder}')
            values_ordered.append(conditions[column])

    diff = len(conditionTexts) - 1 - len(combinations)
    if diff > 0:
        combinations += [ defaultCombination for _ in range(diff) ]

    for index in range(len(conditionTexts)):
        if index != 0:
            text += ' '

        text += f'{conditionTexts[index]}'

        if index < len(conditionTexts)-1:
            text += f' {combinations[index].value}'

    return text, values_ordered

def gen_select(tableName: str, columns: list = [], conditions: dict = {}, combinations: list = [], *, distinct: bool = False, conditionCombining: CombinationTypes = CombinationTypes.AND, conditionComparison: ComparisonTypes = ComparisonTypes.EQUAL_TO, placeholder: str = '?') -> tuple:
    if check_validName(tableName) == False:
        raise exceptions.InvalidTableNameException()

    if len(tableName) == 0:
        raise exceptions.InvalidTableNameException()

    if len(columns) == 0:
        columns = ('*')

    text = 'SELECT '
    if distinct:
        text += 'DISTINCT '

    text += ', '.join(columns)

    text += f' FROM {tableName}'

    column_values_ordered = []

    if len(conditions) > 0:
        text += ' WHERE '

        conditionText, conditionValues = gen_conditions(conditions, combinations=combinations, defaultCombination=conditionCombining, defaultComparison=conditionComparison, placeholder=placeholder)

        text += conditionText
        column_values_ordered += conditionValues

    text += ';'

    return text, column_values_ordered

def gen_count(tableName: str, column: str = "", conditions: dict = {}, combinations: list = [], *, distinct: bool = False, conditionCombining: CombinationTypes = CombinationTypes.AND, conditionComparison: ComparisonTypes = ComparisonTypes.EQUAL_TO, placeholder: str = '?') -> tuple:
    return gen_count_avg_sum(mode=Count_avg_sum_modes.COUNT, tableName=tableName, column=column, conditions=conditions, distinct=distinct, combinations=combinations, conditionCombining=conditionCombining, conditionComparison=conditionComparison, placeholder=placeholder)

def gen_avg(tableName: str, column: str = "", conditions: dict = {}, combinations: list = [], *, distinct: bool = False, conditionCombining: CombinationTypes = CombinationTypes.AND, conditionComparison: ComparisonTypes = ComparisonTypes.EQUAL_TO, placeholder: str = '?') -> tuple:
    return gen_count_avg_sum(mode=Count_avg_sum_modes.AVG, tableName=tableName, column=column, conditions=conditions, distinct=distinct, combinations=combinations, conditionCombining=conditionCombining, conditionComparison=conditionComparison, placeholder=placeholder)

def gen_sum(tableName: str, column: str = "", conditions: dict = {}, combinations: list = [], *, distinct: bool = False, conditionCombining: CombinationTypes = CombinationTypes.AND, conditionComparison: ComparisonTypes = ComparisonTypes.EQUAL_TO, placeholder: str = '?') -> tuple:
    return gen_count_avg_sum(mode=Count_avg_sum_modes.SUM, tableName=tableName, column=column, conditions=conditions, distinct=distinct, combinations=combinations, conditionCombining=conditionCombining, conditionComparison=conditionComparison, placeholder=placeholder)

def gen_count_avg_sum(mode: Count_avg_sum_modes, tableName: str, column: str, conditions: dict = {}, combinations: list = [], *, distinct: bool = False, conditionCombining: CombinationTypes = CombinationTypes.AND, conditionComparison: ComparisonTypes = ComparisonTypes.EQUAL_TO, placeholder: str = '?') -> tuple:
    if not isinstance(mode, Count_avg_sum_modes):
        raise exceptions.InvalidType(mode, Count_avg_sum_modes)

    if check_validName(tableName) == False:
        raise exceptions.InvalidTableNameException()
    elif len(tableName) == 0:
        raise exceptions.InvalidTableNameException()

    if not isinstance(column, str):
        raise exceptions.InvalidType(column, str)

    if len(column) == 0:
        column = "*"

    text = "SELECT "

    if mode == Count_avg_sum_modes.COUNT:
        text += "COUNT"
    elif mode == Count_avg_sum_modes.AVG:
        text += "AVG"
    elif mode == Count_avg_sum_modes.SUM:
        text += "SUM"

    text += '('
    if distinct:
        text += 'DISTINCT '

    text += f"{column}) FROM {tableName}"

    column_values_ordered = []
    if len(conditions) > 0:
        text += ' WHERE '

        conditionText, conditionValues = gen_conditions(conditions, combinations=combinations, defaultCombination=conditionCombining, defaultComparison=conditionComparison, placeholder=placeholder)

        text += conditionText
        column_values_ordered += conditionValues

    text += ';'

    return text, column_values_ordered

def gen_insert(tablename: str, column_values: dict, *, placeholder: str = "?") -> tuple:
    if check_validName(tablename) == False:
        raise exceptions.InvalidTableNameException()

    if len(tablename) == 0:
        raise exceptions.InvalidTableNameException()

    if len(column_values) == 0:
        raise exceptions.DictionaryEmptyException("No columns set.")

    column_values_ordered = []

    text = f'INSERT INTO {tablename} ('

    text += ', '.join([f'{column}' for column in column_values.keys()])

    text += ') VALUES ('

    value_texts = []
    for column in column_values.keys():
        value_texts.append(f'{placeholder}')
        column_values_ordered.append(column_values[column])

    text += ', '.join(value_texts)

    text += ');'

    return text, column_values_ordered

def gen_update(tableName: str, newValues: dict, conditions: dict = {}, combinations: list = [], *, conditionCombining: CombinationTypes = CombinationTypes.AND, conditionComparison: ComparisonTypes = ComparisonTypes.EQUAL_TO, placeholder: str = "?") -> tuple:
    if check_validName(tableName) == False:
        raise exceptions.InvalidTableNameException()

    if len(tableName) == 0:
        raise exceptions.InvalidTableNameException()

    if len(newValues) == 0:
        raise exceptions.DictionaryEmptyException("No values to change specified.")

    column_values_ordered = []

    text = f'UPDATE {tableName} SET '

    setTexts = []
    for column in newValues.keys():
        setText = f'{column}={placeholder}'
        setTexts.append(setText)

        column_values_ordered.append(newValues[column])

    text += ', '.join(setTexts)

    if len(conditions) > 0:
        text += ' WHERE '

        conditionText, conditionValues = gen_conditions(conditions, combinations=combinations, defaultCombination=conditionCombining, defaultComparison=conditionComparison, placeholder=placeholder)

        text += conditionText
        column_values_ordered += conditionValues

    text += ';'

    return text, column_values_ordered

def gen_delete(tableName: str, conditions: dict = {}, combinations: list = [], *, conditionCombining: CombinationTypes = CombinationTypes.AND, conditionComparison: ComparisonTypes = ComparisonTypes.EQUAL_TO, placeholder: str = "?") -> tuple:
    if check_validName(tableName) == False:
        raise exceptions.InvalidTableNameException()

    if len(tableName) == 0:
        raise exceptions.InvalidTableNameException()

    column_values_ordered = []
    text = f'DELETE FROM {tableName}'

    if len(conditions) > 0:
        text += ' WHERE '

        conditionText, conditionValues = gen_conditions(conditions, combinations=combinations, defaultCombination=conditionCombining, defaultComparison=conditionComparison, placeholder=placeholder)

        text += conditionText
        column_values_ordered += conditionValues

    text += ';'

    return text, column_values_ordered

def gen_create_db(dbName: str) -> str:
    if check_validName(dbName) == False:
        raise exceptions.InvalidDatabaseNameException()

    if len(dbName) == 0:
        raise exceptions.InvalidDatabaseNameException()

    text = f'CREATE DATABASE {dbName};'

    return text

def gen_drop_db(dbName: str) -> str:
    if check_validName(dbName) == False:
        raise exceptions.InvalidDatabaseNameException()

    if len(dbName) == 0:
        raise exceptions.InvalidDatabaseNameException()

    text = f'DROP DATABASE {dbName};'

    return text

def gen_create_table(tableName: str, columns: dict, *, primaryKeys: list = [], foreignKeys: dict = {}, namedForeignKeys: dict = {}, uniqueColumns: list = [], safeMode: bool = True) -> str:
    if check_validName(tableName) == False:
        raise exceptions.InvalidTableNameException()

    if len(columns) == 0:
        raise exceptions.DictionaryEmptyException("No columns set.")

    if len(tableName) == 0:
        raise exceptions.InvalidTableNameException()

    for primary in primaryKeys:
        if primary not in columns.keys():
            raise exceptions.InvalidPrimaryKeyColumn(primary)

    for foreign in foreignKeys.keys():
        if foreign not in columns.keys():
            raise exceptions.InvalidForeignKeyColumn(foreign)

    for unique in uniqueColumns:
        if unique not in columns.keys():
            raise exceptions.InvalidUniqueColumn(unique)

    for named in namedForeignKeys.keys():
        if named not in columns.keys():
            raise exceptions.InvalidForeignKeyColumn(named)

        if type(namedForeignKeys[named]) != dict or 'name' not in namedForeignKeys[named].keys() or 'reference' not in namedForeignKeys[named].keys():
            raise exceptions.InvalidNamedForeignKeyDictionary(named)

    text = 'CREATE TABLE '
    if safeMode:
        text += 'IF NOT EXISTS '

    text += f'{tableName} ('

    columnTexts = []

    for column in columns.keys():
        columnText = f'{column} {columns[column]}'

        columnTexts.append(columnText)

    if len(primaryKeys) > 0:
        primaryText = 'PRIMARY KEY ('
        primarySubTexts = []

        for primary in primaryKeys:
            primarySubTexts.append(str(primary))

        primaryText += ','.join(primarySubTexts)
        primaryText += ')'

        columnTexts.append(primaryText)

    for foreign in foreignKeys.keys():
        foreignText = f'FOREIGN KEY ({foreign}) REFERENCES {foreignKeys[foreign]}'

        columnTexts.append(foreignText)

    for foreign in namedForeignKeys.keys():
        foreignText = f'CONSTRAINT {namedForeignKeys[foreign]["name"]} FOREIGN KEY ({foreign}) REFERENCES {namedForeignKeys[foreign]["reference"]}'

        columnTexts.append(foreignText)

    for unique in uniqueColumns:
        uniqueText = f'UNIQUE ({unique})'

        columnTexts.append(uniqueText)

    text += ', '.join(columnTexts)

    text += ');'

    return text

def gen_drop_table(tableName: str) -> str:
    if check_validName(tableName) == False:
        raise exceptions.InvalidTableNameException()

    if len(tableName) == 0:
        raise exceptions.InvalidTableNameException()

    text = f'DROP TABLE {tableName};'

    return text

def gen_create_stored_procedure(procedureName: str, sql_statement: str, parameters: dict = {}) -> str:
    if not isinstance(procedureName, str):
        raise exceptions.InvalidType(procedureName, str)

    if not isinstance(sql_statement, str):
        raise exceptions.InvalidType(sql_statement, str)

    if not isinstance(parameters, dict):
        raise exceptions.InvalidType(parameters, dict)
    else:
        for parameter in parameters:
            if not isinstance(parameter, str) or not isinstance(parameters[parameter], str):
                raise exceptions.InvalidType("Dictionary key and/or value", str)

    text = f"CREATE PROCEDURE {procedureName}"

    if len(parameters) > 0:
        text += ' ' + ', '.join([f'@{key} {parameters[key]}' for key in parameters])

    text += f"\nAS\n{sql_statement}\nGO;"

    return text

def gen_exec_procedure(procedureName: str, parameters: dict = {}, placeholder: str = '?') -> tuple:
    if not isinstance(procedureName, str):
        raise exceptions.InvalidType(procedureName, str)

    if not isinstance(parameters, dict):
        raise exceptions.InvalidType(parameters, dict)

    text = f"EXEC {procedureName}"

    values_ordered = []
    if len(parameters) > 0:
        text += ' '

        paramList = []
        for key in parameters:
            paramList.append(f'@{key}=?')
            values_ordered.append(parameters[key])

        text += ', '.join(paramList)

    text += ";"

    return text, values_ordered

def gen_drop_procedure(procedureName: str) -> str:
    if not isinstance(procedureName, str):
        raise exceptions.InvalidType(procedureName, str)

    return f"DROP PROCEDURE {procedureName};"



from sqlIntuitive.exceptions.baseException import SqlIntuitiveException

class AdaptionProviderException(SqlIntuitiveException):
    def __str__(self):
        return "An overall exception was raised inside the class 'sqlIntuitive.ext.customDataTypes.AdaptionProvider'."



from sqlIntuitive.exceptions.baseException import SqlIntuitiveException

class NotSupported(SqlIntuitiveException):
    def __init__(self, featureName, dbSystemName):
        super().__init__()

        self.featureName = featureName
        self.dbSystemName = dbSystemName

    def __str__(self):
        return f"{self.featureName} is not supported in {self.dbSystemName}."

class CursorIsNone(SqlIntuitiveException):
    def __str__(self):
        return "Cursor was never created."



from base64 import b64encode, b64decode
from sqlIntuitive import exceptions

class CustomDataType():
    def __init__(self, name: str, cls, clsToStringFunc, stringToClsFunc):
        if type(name) != str:
            raise exceptions.NotAString(name)
        self.name = name.upper()

        if isinstance(cls, type) != True:
            raise exceptions.NotAClass(cls)
        self.cls = cls

        if hasattr(clsToStringFunc, '__call__') != True:
            raise exceptions.NotAFunction(clsToStringFunc)
        self.clsToStringFunc = clsToStringFunc

        if hasattr(stringToClsFunc, '__call__') != True:
            raise exceptions.NotAFunction(stringToClsFunc)
        self.stringToClsFunc = stringToClsFunc

    def convertToString(self, clsInstance) -> str:
        if type(clsInstance) != self.cls:
            raise exceptions.NotAMatchingClass(type(clsInstance), self.cls)

        convertedText = self.clsToStringFunc(clsInstance)

        if type(convertedText) != str:
            raise exceptions.NotAString(f"Return type of {self.clsToStringFunc}")

        convertedText = b64encode(convertedText.encode()).decode()

        fullText = f"CUSTOM;{self.name};{convertedText}"

        return fullText

    def convertToClsInstance(self, string: str):
        if type(string) != str:
            raise exceptions.NotAString(string)

        splitted = string.split(';')

        if splitted[1] != self.name:
            return string

        decodedText = b64decode(splitted[2]).decode()

        clsInstance = self.stringToClsFunc(decodedText)
        if isinstance(clsInstance, self.cls) != True:
            raise exceptions.NotAMatchingClass(type(clsInstance), self.cls)

        return clsInstance

class AdaptionProvider():
    def __init__(self):
        self.types = {}
        self.clss = {}

        stringType = StringBypassType("STR_ENCODED")
        self.types[stringType.name] = stringType
        self.clss[stringType.cls] = stringType.name

    def addDataType(self, dataType: CustomDataType) -> None:
        if isinstance(dataType, CustomDataType):
            if dataType.name in self.types.keys() or dataType.cls in self.clss.keys():
                raise exceptions.DuplicationError(dataType.name, dataType.cls)

            self.types[dataType.name] = dataType
            self.clss[dataType.cls] = dataType.name

    def addDataType_raw(self, name: str, cls, clsToStringFunc, stringToClsFunc) -> None:
        dataType = CustomDataType(name, cls, clsToStringFunc, stringToClsFunc)

        self.addDataType(dataType)

    def removeDataType(self, name: str) -> None:
        if type(name) != str:
            raise exceptions.NotAString(name)

        if name.upper() == "STR_ENCODED":
            raise exceptions.DeletingTypeNotAllowed("It is not allowed to remove STR_ENCODED.")

        elif name.upper() not in self.types.keys():
            raise KeyError(f"{name}-type is not set.")

        cls = self.types[name].cls

        del self.clss[cls]
        del self.types[name]

    def convertToString(self, instance):
        if type(instance) in self.clss.keys():
            return self.types[self.clss[type(instance)]].convertToString(instance)
        else:
            return instance

    def convertToClsInstance(self, string: str) -> str:
        if type(string) != str:
            return string

        if string.startswith("CUSTOM;") != True:
            return string

        splitted = string.split(';')

        if splitted[1] not in self.types.keys():
            raise KeyError(f'Type {splitted[1]} is not registered')

        return self.types[splitted[1]].convertToClsInstance(string)

    def convertDictToString(self, dictionary: dict) -> dict:
        if type(dictionary) != dict:
            return dictionary

        dictionary = dictionary.copy()
        for key in dictionary.keys():
            if type(dictionary[key]) == dict:
                if 'value' not in dictionary[key].keys():
                    raise exceptions.NoValue(key)
                dictionary[key]['value'] = self.convertToString(dictionary[key]['value'])
            else:
                dictionary[key] = self.convertToString(dictionary[key])

        return dictionary

    def convertTupleToClsInstance(self, tupleToConvert: tuple) -> tuple:
        if type(tupleToConvert) != tuple:
            return tupleToConvert

        converted = []

        for item in tupleToConvert:
            converted.append(self.convertToClsInstance(item))

        return tuple(converted)

class StringBypassType(CustomDataType):
    def __init__(self, name: str):
        self.name = name
        self.cls = str

    def convertToString(self, clsInstance):
        if clsInstance.startswith("CUSTOM;"):
            encodedText = b64encode(clsInstance.encode()).decode()

            fullText = f"CUSTOM;{self.name};{encodedText}"

            return fullText
        else:
            return clsInstance

    def convertToClsInstance(self, string: str):
        splitted = string.split(';')

        decodedText = b64decode(splitted[2]).decode()

        return decodedText



from sqlIntuitive import sqlGeneration
from sqlIntuitive.ext.customDataTypes import AdaptionProvider, CustomDataType
from sqlIntuitive.conditionEnums import ComparisonTypes, CombinationTypes
from sqlIntuitive.dbSystems.supportTracker import ifSupported, Features, isSupported
from sqlIntuitive.exceptions import CursorIsNone

def cursorNotNone(func):
    def inner(self, *args, **kwargs):
        if self.__getattribute__('cursor') != None:
            return func(self, *args, **kwargs)

        raise CursorIsNone()

    return inner

class BaseDbSystem():
    placeholder = "%s"
    SUPPORTS = Features.ALL.value

    def __init__(self, database: str, adaptionProvider: AdaptionProvider = None) -> None:
        self.database = database

        if isinstance(adaptionProvider, AdaptionProvider):
            self.adaptProvider = adaptionProvider
        else:
            self.adaptProvider = AdaptionProvider()

        self.dbCon = None
        self.cursor = None

    def supports(self, feature: Features) -> bool:
        return isSupported(feature, self.SUPPORTS)

    @ifSupported(Features.ADDON_CUSTOM_DATA_TYPE)
    def addDataType(self, dataType: CustomDataType) -> None:
        self.adaptProvider.addDataType(dataType)

    @ifSupported(Features.ADDON_CUSTOM_DATA_TYPE)
    def addDataType_raw(self, name: str, cls, clsToStringFunc, stringToClsFunc) -> None:
        self.adaptProvider.addDataType_raw(name, cls, clsToStringFunc, stringToClsFunc)

    @ifSupported(Features.SQL_CREATE_TABLE)
    @cursorNotNone
    def create_table(self, tableName: str, columns: dict, *, primaryKeys: list = [], foreignKeys: dict = {}, namedForeignKeys: dict = {}, uniqueColumns: list = [], safeMode: bool = True) -> None:
        sql = sqlGeneration.standard.gen_create_table(tableName, columns, primaryKeys=primaryKeys, foreignKeys=foreignKeys, namedForeignKeys=namedForeignKeys, uniqueColumns=uniqueColumns, safeMode=safeMode)

        self.cursor.execute(sql)

        self.dbCon.commit()

    @ifSupported(Features.SQL_DROP_TABLE)
    @cursorNotNone
    def drop_table(self, tableName: str) -> None:
        sql = sqlGeneration.standard.gen_drop_table(tableName)

        self.cursor.execute(sql)

        self.dbCon.commit()

    @ifSupported(Features.SQL_INSERT_INTO)
    @cursorNotNone
    def insert_into(self, tableName: str, column_values: dict) -> None:
        adaptedColumnValues = self.adaptProvider.convertDictToString(column_values)

        sql, column_values_ordered = sqlGeneration.standard.gen_insert(tableName, adaptedColumnValues, placeholder=self.placeholder)

        self.cursor.execute(sql, column_values_ordered)

        self.dbCon.commit()

    @ifSupported(Features.SQL_UPDATE)
    @cursorNotNone
    def update(self, tableName: str, newColumnValues: dict, conditions: dict = {}, *, conditionCombining: CombinationTypes = CombinationTypes.AND, conditionComparison: ComparisonTypes = ComparisonTypes.EQUAL_TO) -> None:
        adaptedNewColumnValues = self.adaptProvider.convertDictToString(newColumnValues)
        adaptedConditions = self.adaptProvider.convertDictToString(conditions)

        sql, column_values_ordered = sqlGeneration.standard.gen_update(tableName=tableName, newValues=adaptedNewColumnValues, conditions=adaptedConditions, conditionCombining=conditionCombining, conditionComparison=conditionComparison, placeholder=self.placeholder)

        self.cursor.execute(sql, column_values_ordered)

        self.dbCon.commit()

    @ifSupported(Features.SQL_DELETE_FROM)
    @cursorNotNone
    def delete_from(self, tableName: str, conditions: dict = {}, *, conditionCombining: CombinationTypes = CombinationTypes.AND, conditionComparison: ComparisonTypes = ComparisonTypes.EQUAL_TO) -> None:
        adaptedConditions = self.adaptProvider.convertDictToString(conditions)

        sql, column_values_ordered = sqlGeneration.standard.gen_delete(tableName=tableName, conditions=adaptedConditions, conditionCombining=conditionCombining, conditionComparison=conditionComparison, placeholder=self.placeholder)

        self.cursor.execute(sql, column_values_ordered)

        self.dbCon.commit()

    @ifSupported(Features.SQL_SELECT_FROM)
    @cursorNotNone
    def select_from(self, tableName: str, columns: list = [], conditions: dict = {}, *, distinct: bool = False, conditionCombining: CombinationTypes = CombinationTypes.AND, conditionComparison: ComparisonTypes = ComparisonTypes.EQUAL_TO) -> list:
        adaptedConditions = self.adaptProvider.convertDictToString(conditions)

        sql, column_values_ordered = sqlGeneration.standard.gen_select(tableName=tableName, columns=columns, conditions=adaptedConditions, distinct=distinct, conditionCombining=conditionCombining, conditionComparison=conditionComparison, placeholder=self.placeholder)

        self.cursor.execute(sql, column_values_ordered)

        res = self.cursor.fetchall()

        for index in range(len(res)):
            res[index] = self.adaptProvider.convertTupleToClsInstance(res[index])

        return res

    @ifSupported(Features.SQL_COUNT_AVG_SUM)
    def select_count(self, tableName: str, column: str = "", conditions: dict = {}, combinations: list = [], *, distinct: bool = False, conditionCombining: CombinationTypes = CombinationTypes.AND, conditionComparison: ComparisonTypes = ComparisonTypes.EQUAL_TO) -> int:
        return self._select_count_avg_sum(mode=sqlGeneration.standard.Count_avg_sum_modes.COUNT, tableName=tableName, column=column, conditions=conditions, distinct=distinct, combinations=combinations, conditionCombining=conditionCombining, conditionComparison=conditionComparison)

    @ifSupported(Features.SQL_COUNT_AVG_SUM)
    def select_avg(self, tableName: str, column: str = "", conditions: dict = {}, combinations: list = [], *, distinct: bool = False, conditionCombining: CombinationTypes = CombinationTypes.AND, conditionComparison: ComparisonTypes = ComparisonTypes.EQUAL_TO) -> float:
        return self._select_count_avg_sum(mode=sqlGeneration.standard.Count_avg_sum_modes.AVG, tableName=tableName, column=column, conditions=conditions, distinct=distinct, combinations=combinations, conditionCombining=conditionCombining, conditionComparison=conditionComparison)

    @ifSupported(Features.SQL_COUNT_AVG_SUM)
    def select_sum(self, tableName: str, column: str = "", conditions: dict = {}, combinations: list = [], *, distinct: bool = False, conditionCombining: CombinationTypes = CombinationTypes.AND, conditionComparison: ComparisonTypes = ComparisonTypes.EQUAL_TO) -> float:
        return self._select_count_avg_sum(mode=sqlGeneration.standard.Count_avg_sum_modes.SUM, tableName=tableName, column=column, conditions=conditions, distinct=distinct, combinations=combinations, conditionCombining=conditionCombining, conditionComparison=conditionComparison)

    @ifSupported(Features.SQL_COUNT_AVG_SUM)
    @cursorNotNone
    def _select_count_avg_sum(self, mode: sqlGeneration.standard.Count_avg_sum_modes, tableName: str, column: str = "", conditions: dict = {}, combinations: list = [], *, distinct: bool = False, conditionCombining: CombinationTypes = CombinationTypes.AND, conditionComparison: ComparisonTypes = ComparisonTypes.EQUAL_TO):
        adaptedConditions = self.adaptProvider.convertDictToString(conditions)

        sql, column_values_ordered = sqlGeneration.standard.gen_count_avg_sum(mode=mode, tableName=tableName, column=column, conditions=adaptedConditions, distinct=distinct, combinations=combinations, conditionCombining=conditionCombining, conditionComparison=conditionComparison, placeholder=self.placeholder)

        self.cursor.execute(sql, column_values_ordered)

        res = self.cursor.fetchall()

        return res[0][0]

    @ifSupported(Features.SQL_STORED_PROCEDURES)
    @cursorNotNone
    def create_procedure(self, procedureName: str, sql_statement: str, parameters: dict = {}) -> None:
        adaptedParameters = self.adaptionProvider.convertDictToString(parameters)

        sql = sqlGeneration.standard.gen_create_stored_procedure(procedureName=procedureName, sql_statement=sql_statement, parameters=adaptedParameters, placeholder=self.placeholder)

        self.cursor.execute(sql)

    @ifSupported(Features.SQL_STORED_PROCEDURES)
    @cursorNotNone
    def exec_procedure(self,  procedureName: str, parameters: dict = {}) -> list:
        adaptedParameters = self.adaptionProvider.convertDictToString(parameters)

        sql, values_ordered = sqlGeneration.standard.gen_exec_procedure(procedureName=procedureName, parameters=adaptedParameters, placeholder=self.placeholder)

        self.cursor.execute(sql, values_ordered)

        res = self.cursor.fetchall()

        return res

    @ifSupported(Features.SQL_STORED_PROCEDURES)
    @cursorNotNone
    def drop_procedure(self, procedureName: str) -> None:
        sql = sqlGeneration.standard.gen_drop_procedure(procedureName)

        self.cursor.execute(sql)



from sqlIntuitive.exceptions.baseException import SqlIntuitiveException
from sqlIntuitive.exceptions.parameterExceptions import *
from sqlIntuitive.exceptions.conditionExceptions import *
from sqlIntuitive.exceptions.adaptionProviderExceptions import *
from sqlIntuitive.exceptions.other import *



from sqlIntuitive.exceptions.parameterExceptions.baseException import ParameterException

class InvalidConstraint(ParameterException):
    def __str__(self):
        return "An overall \'Invalid constraint\' parameter exception was raised."



from sqlIntuitive.ext import customDataTypes
from sqlIntuitive import exceptions, conditionEnums

import unittest

import json, os

defaultFile = 'tests/testsToRun.json'
altFile = 'tests/testsToRunDefault.json'

with open((defaultFile if os.path.exists(defaultFile) else altFile), 'r') as file:
    runTestCustomDataTypes = json.load(file)['customDataTypes']

@unittest.skipIf(runTestCustomDataTypes == False, 'Skipped TestDBSystems via config')
class TestcustomDataTypes(unittest.TestCase):
    def setUpClass():
        cls = TestcustomDataTypes

        class Test():
            pass

        def clsToString(clsInstance):
            return str(clsInstance)

        def stringToCls(string):
            return Test()

        cls.Test = Test
        cls.clsToString = clsToString
        cls.stringToCls = stringToCls

        class TestB():
            def __init__(self, num):
                self.num = num

        def testBToString(instance):
            return str(instance.num)

        def stringToTestB(string):
            return TestB(int(string))

        cls.TestB = TestB
        cls.testBToString = testBToString
        cls.stringToTestB = stringToTestB

        class MyClass():
            def __init__(self, arg):
                self.arg = arg

        def myClassToString(clsInstance):
            return str(clsInstance.arg)

        def stringToMyClass(string):
            return MyClass(string)

        cls.MyClass = MyClass
        cls.myClassToString = myClassToString
        cls.stringToMyClass = stringToMyClass

    def test_A_customDataType_init(self):
        customType = customDataTypes.CustomDataType("Test", TestcustomDataTypes.Test, TestcustomDataTypes.clsToString, TestcustomDataTypes.stringToCls)

        self.assertEqual(customType.name, "TEST")

    def test_B_customDataType_convertToString(self):
        customType = customDataTypes.CustomDataType("Test", TestcustomDataTypes.TestB, TestcustomDataTypes.testBToString, TestcustomDataTypes.stringToTestB)

        self.assertEqual(customType.convertToString( TestcustomDataTypes.TestB(123) ), "CUSTOM;TEST;MTIz")

        instance = customType.convertToClsInstance("CUSTOM;TEST;MTIz")
        self.assertTrue(isinstance(instance, TestcustomDataTypes.TestB))
        self.assertEqual(instance.num, 123)

    def test_C_adaptionProvider_normalValues(self):
        adaptProvider = customDataTypes.AdaptionProvider()

        self.assertEqual(adaptProvider.convertToString('this is a test string'), 'this is a test string')
        self.assertEqual(adaptProvider.convertToClsInstance('this is a test string'), 'this is a test string')

        self.assertEqual(adaptProvider.convertToString(123), 123)
        self.assertEqual(adaptProvider.convertToClsInstance(123), 123)

    def test_D_adaptionProvider_str_encoded(self):
        adaptProvider = customDataTypes.AdaptionProvider()

        self.assertEqual(adaptProvider.convertToString("CUSTOM;4v67q8936mnpq2"), "CUSTOM;STR_ENCODED;Q1VTVE9NOzR2NjdxODkzNm1ucHEy")
        self.assertEqual(adaptProvider.convertToClsInstance("CUSTOM;STR_ENCODED;Q1VTVE9NOzR2NjdxODkzNm1ucHEy"), "CUSTOM;4v67q8936mnpq2")

    def test_E_adaptionProvider_customType(self):
        adaptProvider = customDataTypes.AdaptionProvider()

        dataType = customDataTypes.CustomDataType("MYCLS", TestcustomDataTypes.MyClass, TestcustomDataTypes.myClassToString, TestcustomDataTypes.stringToMyClass)
        adaptProvider.addDataType(dataType)

        self.assertEqual(adaptProvider.convertToString(TestcustomDataTypes.MyClass("abc")), "CUSTOM;MYCLS;YWJj")

        clsInstance = adaptProvider.convertToClsInstance("CUSTOM;MYCLS;YWJj")
        self.assertEqual(type(clsInstance), TestcustomDataTypes.MyClass)
        self.assertEqual(clsInstance.arg, "abc")

    def test_F_adaptionProvicer_raw_add(self):
        adaptProvider = customDataTypes.AdaptionProvider()

        adaptProvider.addDataType_raw("MYCLS", TestcustomDataTypes.MyClass, TestcustomDataTypes.myClassToString, TestcustomDataTypes.stringToMyClass)

        self.assertTrue("MYCLS" in adaptProvider.types)

    def test_G_adaptionProvider_removeDataType(self):
        adaptProvider = customDataTypes.AdaptionProvider()

        dataType = customDataTypes.CustomDataType("MYCLS", TestcustomDataTypes.MyClass, TestcustomDataTypes.myClassToString, TestcustomDataTypes.stringToMyClass)
        adaptProvider.addDataType(dataType)

        self.assertTrue("MYCLS" in adaptProvider.types)

        adaptProvider.removeDataType(dataType.name)

        self.assertFalse("MYCLS" in adaptProvider.types, {})

    def test_H_customDataType_error_test(self):
        with self.assertRaises(exceptions.NotAString):
            dataType = customDataTypes.CustomDataType(123, TestcustomDataTypes.MyClass, TestcustomDataTypes.myClassToString, TestcustomDataTypes.stringToMyClass)

        with self.assertRaises(exceptions.NotAClass):
            dataType = customDataTypes.CustomDataType("MYCLS", "abc", TestcustomDataTypes.myClassToString, TestcustomDataTypes.stringToMyClass)

        with self.assertRaises(exceptions.NotAFunction):
            dataType = customDataTypes.CustomDataType("MYCLS", TestcustomDataTypes.MyClass, None, TestcustomDataTypes.stringToMyClass)

        with self.assertRaises(exceptions.NotAFunction):
            dataType = customDataTypes.CustomDataType("MYCLS", TestcustomDataTypes.MyClass, TestcustomDataTypes.myClassToString, None)

        dataType = customDataTypes.CustomDataType("MYCLS", TestcustomDataTypes.MyClass, TestcustomDataTypes.myClassToString, TestcustomDataTypes.stringToMyClass)
        with self.assertRaises(exceptions.NotAMatchingClass):
            dataType.convertToString(TestcustomDataTypes.TestB(123))

        def wrongConvertToString(clsInstance):
            return clsInstance

        dataType = customDataTypes.CustomDataType("MYCLS", TestcustomDataTypes.MyClass, wrongConvertToString, TestcustomDataTypes.stringToMyClass)
        with self.assertRaises(exceptions.NotAString):
            dataType.convertToString(TestcustomDataTypes.MyClass(123))

        dataType = customDataTypes.CustomDataType("MYCLS", TestcustomDataTypes.MyClass, TestcustomDataTypes.myClassToString, TestcustomDataTypes.stringToMyClass)
        with self.assertRaises(exceptions.NotAString):
            dataType.convertToClsInstance(123)

        def wrongConvertToClsInstance(string):
            return string

        dataType = customDataTypes.CustomDataType("MYCLS", TestcustomDataTypes.MyClass, TestcustomDataTypes.myClassToString, wrongConvertToClsInstance)
        with self.assertRaises(exceptions.NotAMatchingClass):
            dataType.convertToClsInstance("CUSTOM;MYCLS;MTIz")

    def test_I_adaptionProvider_error_test(self):
        adaptProvider = customDataTypes.AdaptionProvider()
        dataType = customDataTypes.CustomDataType("MYCLS", TestcustomDataTypes.MyClass, TestcustomDataTypes.myClassToString, TestcustomDataTypes.stringToMyClass)
        dataType2 = customDataTypes.CustomDataType("MYCLS", TestcustomDataTypes.MyClass, TestcustomDataTypes.myClassToString, TestcustomDataTypes.stringToMyClass)
        adaptProvider.addDataType(dataType)

        with self.assertRaises(exceptions.DuplicationError):
            adaptProvider.addDataType(dataType2)

        adaptProvider = customDataTypes.AdaptionProvider()
        with self.assertRaises(exceptions.NotAString):
            adaptProvider.removeDataType(123)

        adaptProvider = customDataTypes.AdaptionProvider()
        with self.assertRaises(exceptions.DeletingTypeNotAllowed):
            adaptProvider.removeDataType("STR_ENCODED")

        adaptProvider = customDataTypes.AdaptionProvider()
        with self.assertRaises(KeyError):
            adaptProvider.removeDataType("ABC")

        adaptProvider = customDataTypes.AdaptionProvider()
        with self.assertRaises(exceptions.NoValue):
            adaptProvider.convertDictToString({'col1': {}})

    def test_J_convertTupleToClsInstance(self):
        adaptProvider = customDataTypes.AdaptionProvider()
        dataType = customDataTypes.CustomDataType("MYCLS", TestcustomDataTypes.MyClass, TestcustomDataTypes.myClassToString, TestcustomDataTypes.stringToMyClass)

        adaptProvider.addDataType(dataType)

        testTuple = ('CUSTOM;MYCLS;VGVzdA==', 123, "CUSTOM;STR_ENCODED;Q1VTVE9NO2JsYWJsYWJsYQ==")

        converted = adaptProvider.convertTupleToClsInstance(testTuple)

        self.assertTrue(isinstance(converted[0], TestcustomDataTypes.MyClass))
        self.assertEqual(converted[0].arg, 'Test')

        self.assertEqual(converted[1], 123)
        self.assertEqual(converted[2], 'CUSTOM;blablabla')

    def test_K_convertDictToString(self):
        adaptProvider = customDataTypes.AdaptionProvider()

        testDict = {'col1': 123, 'col2': 'abc', 'col3': 'CUSTOM;anythinggoeshere'}

        converted = adaptProvider.convertDictToString(testDict)

        expected = {'col1': 123, 'col2': 'abc', 'col3': 'CUSTOM;STR_ENCODED;Q1VTVE9NO2FueXRoaW5nZ29lc2hlcmU='}

        self.assertEqual(converted, expected)

    def test_L_convertDictToString_customDataType(self):
        adaptProvider = customDataTypes.AdaptionProvider()
        dataType = customDataTypes.CustomDataType("MYCLS", TestcustomDataTypes.MyClass, TestcustomDataTypes.myClassToString, TestcustomDataTypes.stringToMyClass)

        adaptProvider.addDataType(dataType)

        clsInstance = TestcustomDataTypes.MyClass('Test')
        testDict = {'col1': {'value': clsInstance}, 'col2': {'value': 123, 'comparison': conditionEnums.ComparisonTypes.EQUAL_TO}}

        converted = adaptProvider.convertDictToString(testDict)

        expected =  {'col1': {'value': 'CUSTOM;MYCLS;VGVzdA=='}, 'col2': {'value': 123, 'comparison': conditionEnums.ComparisonTypes.EQUAL_TO}}

        self.assertEqual(converted, expected)



from sqlIntuitive.exceptions.adaptionProviderExceptions.baseException import AdaptionProviderException

class DuplicationError(AdaptionProviderException):
    def __init__(self, name, cls):
        super().__init__()

        self.name = name
        self.cls = cls

    def __str__(self):
        return f"{self.name} ({self.cls}) are already registered."


class DeletingTypeNotAllowed(AdaptionProviderException):
    def __init__(self, name):
        super().__init__()

        self.name = name

    def __str__(self):
        return f"It is not allowed to remove {self.name}."



from enum import Enum

class ComparisonTypes(Enum):
    EQUAL_TO = "="
    GREATER_THAN = ">"
    LESS_THAN = "<"
    GREATER_THAN_OR_EQUAL_TO = ">="
    LESS_THAN_OR_EQUAL_TO = "<="
    NOT_EQUAL_TO = "!="

class CombinationTypes(Enum):
    AND = "AND"
    OR = "OR"



from setuptools import find_packages, setup

import re
import json

with open('requirements.txt', 'r') as file:
    requirements = [re.sub('\n','',i) for i in file]

with open('version.json', 'r') as file:
    version = json.load(file)['version']

setup(
    name='sqlIntuitive',
    python_requires='>=3.8',
    package_dir={"":"src"},
    packages=find_packages(where="src"),
    include_package_data=True,
    version=version,
    description='Library for easy and intuitive use of database systems.',
    long_description='Library for easy and intuitive use of database systems.',
    author="Moritz Bauer",
    url="https://www.github.com/einfachIrgendwer0815/SqlIntuitive",
    license="MIT",
    install_requires=requirements,
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
    test_suite='tests'
)



from sqlIntuitive.exceptions.conditionExceptions.baseException import ConditionException

class NotAComparisonType(ConditionException):
    def __init__(self, givenType):
        super().__init__()

        self.type = givenType

    def __str__(self):
        return f"{self.type} is not a comparison type."

class NotACombinationType(ConditionException):
    def __init__(self, givenType):
        super().__init__()

        self.type = givenType

    def __str__(self):
        return f"{self.type} is not a combination type."

class NoValue(ConditionException):
    def __init__(self, key):
        super().__init__()

        self.key = key

    def __str__(self):
        return f"No value for {self.key}."



from sqlIntuitive import dbSystems
from sqlIntuitive.ext import customDataTypes

from sqlIntuitive import exceptions

from sqlIntuitive.conditionEnums import ComparisonTypes, CombinationTypes

import sqlite3
import unittest

import json, os

defaultFile = 'tests/testsToRun.json'
altFile = 'tests/testsToRunDefault.json'

dbFile = "tests/test_D.db.sqlite3"

with open((defaultFile if os.path.exists(defaultFile) else altFile), 'r') as file:
    runTestSqliteSystem = json.load(file)['sqliteSystem']

@unittest.skipIf(runTestSqliteSystem == False, 'Skipped TestSqliteSystem via config')
class TestSqliteSystem(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        file_db = sqlite3.connect(dbFile)

        cursor = file_db.cursor()

        cursor.execute("CREATE TABLE IF NOT EXISTS TableA (col1 string, col2 int, col3 bool, col4 string, PRIMARY KEY (col1), UNIQUE (col4));")

        cursor.close()
        file_db.close()

    @classmethod
    def tearDownClass(cls):
        os.remove(dbFile)

    def setUp(self):
        self.memory_db = dbSystems.SqliteDbSystem(database=":memory:")
        self.file_db = dbSystems.SqliteDbSystem(database=dbFile)

        self.memory_db.connect_to_db()
        self.file_db.connect_to_db()

    def tearDown(self):
        self.memory_db.close_connection()
        self.file_db.close_connection()

    def test_A_connect_to_db(self):
        self.assertIsNotNone(self.memory_db.dbCon)

    def test_B_cursor(self):
        self.assertIsNone(self.memory_db.cursor)

        self.memory_db.create_cursor()

        self.assertIsNotNone(self.memory_db.cursor)

    def test_C_adaptionProvider(self):
        self.assertTrue(isinstance(self.memory_db.adaptProvider, customDataTypes.AdaptionProvider))

        adaptProvider = customDataTypes.AdaptionProvider()
        memory_db2 = dbSystems.SqliteDbSystem(":memory:", adaptionProvider=adaptProvider)

        self.assertEqual(memory_db2.adaptProvider, adaptProvider)

    def test_D_open_close(self):
        self.memory_db.create_cursor() # Won't work but no error is raised

        self.memory_db.connect_to_db()
        self.memory_db.create_cursor()
        self.memory_db.close_connection()

        self.memory_db.create_cursor() # Won't work too but still no error is raised

    def test_E_create_table(self):
        self.file_db.create_cursor()

        self.file_db.create_table("TableB", {"name":"varchar(50)","id": "int", "id2": "int", "col3": 'int', "col4": "varchar(50)", "col5": "int"}, primaryKeys=['id', 'id2'], foreignKeys={'col3': 'TableB(col2)'}, namedForeignKeys={'col5': {'name':'myTestFK', 'reference': "TableB(col2)"}}, uniqueColumns=['col4'])

    def test_F_drop_table(self):
        self.file_db.create_cursor()

        self.file_db.drop_table("TableB")

    def test_G_insert_into(self):
        self.file_db.create_cursor()

        self.file_db.insert_into("TableA", {"col1": 'Test', "col2": 42, "col3": True})
        self.file_db.insert_into("TableA", {"col1": 'Test2', "col2": 84, "col3": True})


    def test_H_update(self):
        self.file_db.create_cursor()

        self.file_db.update("TableA", {"col3": False, "col2": 43}, {"col2": 42})
        self.file_db.update("TableA", {"col3": False, "col2": 44}, {"col2": {'value': 43, 'comparison': ComparisonTypes.LESS_THAN_OR_EQUAL_TO}})

    def test_I_select_from(self):
        self.memory_db.create_cursor()

        self.memory_db.create_table("TableA", {'col1': 'string', 'col2': 'int'})
        self.memory_db.insert_into("TableA", {'col1': 'ABC', 'col2': 123})
        self.memory_db.insert_into("TableA", {'col1': 'DEF', 'col2': 456})
        self.memory_db.insert_into("TableA", {'col1': 'ABC', 'col2': 789})
        self.memory_db.insert_into("TableA", {'col1': 'GHI', 'col2': 123})

        res = self.memory_db.select_from("TableA")
        self.assertEqual(res, [('ABC',123), ('DEF',456), ('ABC',789),('GHI',123)])

        res = self.memory_db.select_from("TableA", conditions={'col2': {'value':456, 'comparison': ComparisonTypes.LESS_THAN_OR_EQUAL_TO}})
        self.assertEqual(res, [('ABC',123),('DEF',456), ('GHI',123)])

        res = self.memory_db.select_from("TableA", ['col2'], distinct=True)
        self.assertEqual(res, [(123,),(456,),(789,)])

    def test_J_delete(self):
        self.file_db.create_cursor()

        self.file_db.delete_from("TableA", {"col2": 84})
        self.file_db.delete_from("TableA", {"col2": {'value': 80, 'comparison': ComparisonTypes.EQUAL_TO}})
        self.file_db.delete_from("TableA")

    def test_K_str_encoded(self):
        self.assertTrue(isinstance(self.file_db.adaptProvider, customDataTypes.AdaptionProvider))

        self.file_db.create_cursor()

        self.file_db.insert_into("TableA", {'col1': "CUSTOM;", 'col2': 6473573})

        res = self.file_db.select_from("TableA", columns=['col1'], conditions={'col2': 6473573})
        self.assertEqual(res, [("CUSTOM;",)])

    def test_L_customDataType(self):
        self.assertTrue(isinstance(self.file_db.adaptProvider, customDataTypes.AdaptionProvider))

        class TestClass():
            def __init__(self, num):
                self.num = num

        def testClassToString(instance):
            return str(instance.num)

        def stringToTestClass(string):
            return TestClass(int(string))

        dataType = customDataTypes.CustomDataType("TESTCLS", TestClass, testClassToString, stringToTestClass)

        self.file_db.addDataType(dataType)

        self.assertTrue('TESTCLS' in self.file_db.adaptProvider.types)

        self.file_db.create_cursor()

        self.file_db.insert_into("TableA", {'col1': TestClass(123), 'col2': 567})
        res = self.file_db.select_from("TableA", conditions={'col2': 567})

        self.assertTrue(isinstance(res[0][0], TestClass))

    def test_M_select_count(self):
        self.memory_db.create_cursor()

        self.memory_db.create_table("TableB", {"col1": "string", "col2": "int"})
        self.memory_db.insert_into("TableB", {"col1": "ABC", "col2": 123})
        self.memory_db.insert_into("TableB", {"col1": "DEF", "col2": 456})
        self.memory_db.insert_into("TableB", {"col1": "ABC", "col2": 789})

        res = self.memory_db.select_count("TableB", column="col2")
        self.assertEqual(res, 3)

        res = self.memory_db.select_count("TableB", column="col2", conditions={"col1": "ABC"})
        self.assertEqual(res, 2)

        res = self.memory_db.select_count("TableB", column="col2", conditions={"col2": {"value": 123, "comparison": ComparisonTypes.GREATER_THAN}})
        self.assertEqual(res, 2)


        res = self.memory_db.select_count("TableB", column="col1")
        self.assertEqual(res,3)

        res = self.memory_db.select_count("TableB", column="col1", distinct=True)
        self.assertEqual(res,2)

    def test_N_select_avg(self):
        self.memory_db.create_cursor()

        self.memory_db.create_table("TableB", {"col1": "string", "col2": "int"})
        self.memory_db.insert_into("TableB", {"col1": "ABC", "col2": 123})
        self.memory_db.insert_into("TableB", {"col1": "DEF", "col2": 456})
        self.memory_db.insert_into("TableB", {"col1": "ABC", "col2": 789})

        res = self.memory_db.select_avg("TableB", column="col2")
        self.assertEqual(res, 456)

        res = self.memory_db.select_avg("TableB", column="col2", conditions={"col1": "ABC"})
        self.assertEqual(res, (123+789)/2)

        res = self.memory_db.select_avg("TableB", column="col2", conditions={"col2": {"value": 123, "comparison": ComparisonTypes.GREATER_THAN}})
        self.assertEqual(res, (456+789)/2)


        self.memory_db.insert_into("TableB", {"col1": "GHI", "col2": 123})

        res = self.memory_db.select_avg("TableB", column="col2")
        self.assertEqual(res, (123+456+789+123)/4)

        res = self.memory_db.select_avg("TableB", column="col2", distinct=True)
        self.assertEqual(res, (123+456+789)/3)

    def test_O_select_sum(self):
        self.memory_db.create_cursor()

        self.memory_db.create_table("TableB", {"col1": "string", "col2": "int"})
        self.memory_db.insert_into("TableB", {"col1": "ABC", "col2": 123})
        self.memory_db.insert_into("TableB", {"col1": "DEF", "col2": 456})
        self.memory_db.insert_into("TableB", {"col1": "ABC", "col2": 789})

        res = self.memory_db.select_sum("TableB", column="col2")
        self.assertEqual(res, 123+456+789)

        res = self.memory_db.select_sum("TableB", column="col2", conditions={"col1": "ABC"})
        self.assertEqual(res, 123+789)

        res = self.memory_db.select_sum("TableB", column="col2", conditions={"col2": {"value": 123, "comparison": ComparisonTypes.GREATER_THAN}})
        self.assertEqual(res, 456+789)


        self.memory_db.insert_into("TableB", {"col1": "GHI", "col2": 123})

        res = self.memory_db.select_sum("TableB", column="col2")
        self.assertEqual(res, 123+456+789+123)

        res = self.memory_db.select_sum("TableB", column="col2", distinct=True)
        self.assertEqual(res, 123+456+789)

    def test_P_cursorIsNone(self):
        with self.assertRaises(exceptions.CursorIsNone):
            self.memory_db.create_table("TableB", {"col1": "string"})

        with self.assertRaises(exceptions.CursorIsNone):
            self.memory_db.drop_table("TableB")

        with self.assertRaises(exceptions.CursorIsNone):
            self.memory_db.insert_into("TableB", {"col1": "abc"})

        with self.assertRaises(exceptions.CursorIsNone):
            self.memory_db.update("TableB", {"col1": "abc"})

        with self.assertRaises(exceptions.CursorIsNone):
            self.memory_db.delete_from("TableB")

        with self.assertRaises(exceptions.CursorIsNone):
            self.memory_db.select_from("TableB")

        with self.assertRaises(exceptions.CursorIsNone):
            self.memory_db.select_count("TableB", "col1")

        with self.assertRaises(exceptions.CursorIsNone):
            self.memory_db.select_avg("TableB", "col1")

        with self.assertRaises(exceptions.CursorIsNone):
            self.memory_db.select_sum("TableB", "col1")



from sqlIntuitive import exceptions
from sqlIntuitive.sqlGeneration import standard
from sqlIntuitive.conditionEnums import ComparisonTypes, CombinationTypes

import unittest

import random, string

import json, os

defaultFile = 'tests/testsToRun.json'
altFile = 'tests/testsToRunDefault.json'

with open((defaultFile if os.path.exists(defaultFile) else altFile), 'r') as file:
    runTestSqlGenerationStandard = json.load(file)['sqlGeneration_standard']

@unittest.skipIf(runTestSqlGenerationStandard == False, 'Skipped TestSqlGenerationStandard via config')
class TestSqlGenerationStandard(unittest.TestCase):
    def test_A_A_a_gen_conditions(self):
        self.assertEqual(standard.gen_conditions({'abc': 'def', 'xyz': {'value': 123}, 'test': {'value': 42, 'comparison': ComparisonTypes.GREATER_THAN}}), ('abc=? AND xyz=? AND test>?', ['def', 123, 42]))
        self.assertEqual(standard.gen_conditions({'abc': 'def', 'xyz': {'value': 123}, 'test': {'value': 42, 'comparison': ComparisonTypes.GREATER_THAN}}, defaultComparison=ComparisonTypes.LESS_THAN_OR_EQUAL_TO), ('abc<=? AND xyz<=? AND test>?', ['def', 123, 42]))
        self.assertEqual(standard.gen_conditions({'abc': 'def', 'xyz': {'value': 123}, 'test': {'value': 42, 'comparison': ComparisonTypes.GREATER_THAN}}, defaultCombination=CombinationTypes.OR), ('abc=? OR xyz=? OR test>?', ['def', 123, 42]))

        self.assertEqual(standard.gen_conditions({'abc': 'def', 'xyz': {'value': 123}, 'test': {'value': 42, 'comparison': ComparisonTypes.GREATER_THAN}}, combinations=[CombinationTypes.OR]), ('abc=? OR xyz=? AND test>?', ['def', 123, 42]))
        self.assertEqual(standard.gen_conditions({'abc': 'def', 'xyz': {'value': 123}, 'test': {'value': 42, 'comparison': ComparisonTypes.GREATER_THAN}}, combinations=[CombinationTypes.OR, CombinationTypes.AND, CombinationTypes.OR]), ('abc=? OR xyz=? AND test>?', ['def', 123, 42]))

        self.assertEqual(standard.gen_conditions({'abc': 'def', 'xyz': {'value': 123}, 'test': {'value': 42, 'comparison': ComparisonTypes.GREATER_THAN}}, defaultCombination=CombinationTypes.OR), ('abc=? OR xyz=? OR test>?', ['def', 123, 42]))
        self.assertEqual(standard.gen_conditions({'abc': 'def', 'xyz': {'value': 123}, 'test': {'value': 42, 'comparison': ComparisonTypes.GREATER_THAN}}, combinations=[CombinationTypes.AND], defaultCombination=CombinationTypes.OR), ('abc=? AND xyz=? OR test>?', ['def', 123, 42]))

        self.assertEqual(standard.gen_conditions(), ('',[]))
        self.assertEqual(standard.gen_conditions({'abc':'def', 'x':'y'}, defaultCombination="OR"), ('abc=? OR x=?',['def','y']))
        self.assertEqual(standard.gen_conditions({'abc':'def', 'x':'y'}, defaultComparison=">="), ('abc>=? AND x>=?',['def','y']))
        self.assertEqual(standard.gen_conditions({'abc':'def', 'x':{'value': 'y', 'comparison': ">="}}), ('abc=? AND x>=?',['def','y']))
        self.assertEqual(standard.gen_conditions({'abc':'def', 'x':'y', 'hello': 'world'}, combinations=["OR", "AND"]), ('abc=? OR x=? AND hello=?',['def','y','world']))

    def test_A_A_b_gen_conditions(self):
        with self.assertRaises(exceptions.InvalidType):
            standard.gen_conditions(['test'])

        with self.assertRaises(exceptions.InvalidType):
            standard.gen_conditions(combinations={})

        with self.assertRaises(exceptions.NotACombinationType):
            standard.gen_conditions(defaultCombination='abc')

        with self.assertRaises(exceptions.NotACombinationType):
            standard.gen_conditions({}, combinations=[CombinationTypes.AND, 'xyz'])

        with self.assertRaises(exceptions.NotAComparisonType):
            standard.gen_conditions(defaultComparison="abc")

        with self.assertRaises(exceptions.NotAComparisonType):
            standard.gen_conditions({'abc': {'value': 'def', 'comparison': 'xy'}})

        with self.assertRaises(exceptions.NoValue):
            standard.gen_conditions({'abc': {}})

        with self.assertRaises(exceptions.NoValue):
            standard.gen_conditions({'abc': {'comparison': ComparisonTypes.EQUAL_TO}})

    def test_A_A_c_gen_insert(self):
        self.assertEqual(standard.gen_insert("TestA", {"colA":"val1", "colB": 123, "colC": True}), ("INSERT INTO TestA (colA, colB, colC) VALUES (?, ?, ?);", ['val1', 123, True]))

    def test_A_B_gen_insert(self):
        with self.assertRaises(exceptions.DictionaryEmptyException):
            standard.gen_insert("TestB", {})

        with self.assertRaises(exceptions.InvalidTableNameException):
            standard.gen_insert(" Test B", {"colA":"val1", "colB": 123, "colC": True})

    def test_A_C_gen_insert(self):
        with self.assertRaises(exceptions.InvalidTableNameException):
            standard.gen_insert("", {"colA": "val1"})

        with self.assertRaises(exceptions.InvalidTableNameException):
            standard.gen_insert(" ", {"colA": "val1"})

    def test_A_D_gen_delete(self):
        self.assertEqual(standard.gen_delete("TestD", {"colXY": "ABC", "id": 12345}), ("DELETE FROM TestD WHERE colXY=? AND id=?;", ['ABC', 12345]))
        self.assertEqual(standard.gen_delete("TestD", {"colXY": "ABC", "id": 12345}, conditionCombining='OR'), ("DELETE FROM TestD WHERE colXY=? OR id=?;", ['ABC', 12345]))
        self.assertEqual(standard.gen_delete("TestD"), ("DELETE FROM TestD;", []))

    def test_A_E_gen_delete(self):
        with self.assertRaises(exceptions.InvalidTableNameException):
            standard.gen_delete("")

        with self.assertRaises(exceptions.InvalidTableNameException):
            standard.gen_delete("", {"col": "val"})

        with self.assertRaises(exceptions.InvalidTableNameException):
            standard.gen_delete("", {"col": "val"}, "OR")

        with self.assertRaises(exceptions.InvalidTableNameException):
            standard.gen_delete(" ")

        with self.assertRaises(exceptions.InvalidTableNameException):
            standard.gen_delete(" ", {"col": "val"})

        with self.assertRaises(exceptions.InvalidTableNameException):
            standard.gen_delete(" ", {"col": "val"}, "OR")

        with self.assertRaises(exceptions.InvalidTableNameException):
            standard.gen_delete("TestD ")

        with self.assertRaises(exceptions.InvalidTableNameException):
            standard.gen_delete("T es tD", {"colXY": "ABC", "id": 12345})

        with self.assertRaises(exceptions.InvalidTableNameException):
            standard.gen_delete("T es tD", {"colXY": "ABC", "id": 12345}, 'OR')

        with self.assertRaises(exceptions.InvalidTableNameException):
            standard.gen_delete("T es tD")

        with self.assertRaises(exceptions.InvalidTableNameException):
            standard.gen_delete("T es tD ")

    def test_A_F_gen_create_db(self):
        self.assertTrue(standard.gen_create_db("TestF"))

    def test_A_G_gen_create_db(self):
        with self.assertRaises(exceptions.InvalidDatabaseNameException):
            standard.gen_create_db("")

        with self.assertRaises(exceptions.InvalidDatabaseNameException):
            standard.gen_create_db(" ")

        with self.assertRaises(exceptions.InvalidDatabaseNameException):
            standard.gen_create_db("   ")

        with self.assertRaises(exceptions.InvalidDatabaseNameException):
            standard.gen_create_db(" Tes tF ")

        with self.assertRaises(exceptions.InvalidDatabaseNameException):
            standard.gen_create_db(" TestF")

        with self.assertRaises(exceptions.InvalidDatabaseNameException):
            standard.gen_create_db("TestF ")

        with self.assertRaises(exceptions.InvalidDatabaseNameException):
            standard.gen_create_db(" Tes   tF ")

    def test_A_H_gen_create_table(self):
        self.assertEqual(standard.gen_create_table("TestH", {"col1": "int"}, safeMode=False), "CREATE TABLE TestH (col1 int);")
        self.assertEqual(standard.gen_create_table("TestH", {"col1": "int"}, safeMode=True), "CREATE TABLE IF NOT EXISTS TestH (col1 int);")
        self.assertEqual(standard.gen_create_table("TestH", {"col1": "int"}), "CREATE TABLE IF NOT EXISTS TestH (col1 int);")
        self.assertEqual(standard.gen_create_table("TestH", {"col1": "int", "xyz": "char"}, safeMode=False), "CREATE TABLE TestH (col1 int, xyz char);")
        self.assertEqual(standard.gen_create_table("TestH", {"col1": "int", "xyz": "char"}, safeMode=True), "CREATE TABLE IF NOT EXISTS TestH (col1 int, xyz char);")
        self.assertEqual(standard.gen_create_table("TestH", {"col1": "int", "xyz": "char"}), "CREATE TABLE IF NOT EXISTS TestH (col1 int, xyz char);")

        self.assertEqual(standard.gen_create_table("TestH", {"col1": "int", "xyz": "char"}, primaryKeys=['col1']), "CREATE TABLE IF NOT EXISTS TestH (col1 int, xyz char, PRIMARY KEY (col1));")
        self.assertEqual(standard.gen_create_table("TestH", {"col1": "int", "xyz": "char"}, primaryKeys=['col1', 'xyz']), "CREATE TABLE IF NOT EXISTS TestH (col1 int, xyz char, PRIMARY KEY (col1,xyz));")
        self.assertEqual(standard.gen_create_table("TestH", {"col1": "int", "xyz": "char"}, foreignKeys={'col1': 'Test(Hello)'}), "CREATE TABLE IF NOT EXISTS TestH (col1 int, xyz char, FOREIGN KEY (col1) REFERENCES Test(Hello));")
        self.assertEqual(standard.gen_create_table("TestH", {"col1": "int", "xyz": "char"}, foreignKeys={'col1': 'Test(Hello)', 'xyz': 'Test(World)'}), "CREATE TABLE IF NOT EXISTS TestH (col1 int, xyz char, FOREIGN KEY (col1) REFERENCES Test(Hello), FOREIGN KEY (xyz) REFERENCES Test(World));")

        self.assertEqual(standard.gen_create_table("TestH", {"col1": "int", "xyz": "char"}, primaryKeys=['xyz'], foreignKeys={'col1': 'Test(Hello)'}), "CREATE TABLE IF NOT EXISTS TestH (col1 int, xyz char, PRIMARY KEY (xyz), FOREIGN KEY (col1) REFERENCES Test(Hello));")

        self.assertEqual(standard.gen_create_table("TestH", {"col1": "int", "xyz": "char"}, uniqueColumns=['col1']), "CREATE TABLE IF NOT EXISTS TestH (col1 int, xyz char, UNIQUE (col1));")
        self.assertEqual(standard.gen_create_table("TestH", {"col1": "int", "xyz": "char", "abc": "int"}, primaryKeys=['xyz'], foreignKeys={'col1': 'Test(Hello)'}, uniqueColumns=['abc']), "CREATE TABLE IF NOT EXISTS TestH (col1 int, xyz char, abc int, PRIMARY KEY (xyz), FOREIGN KEY (col1) REFERENCES Test(Hello), UNIQUE (abc));")

        self.assertEqual(standard.gen_create_table("TestH", {"col1": "varchar(10)"}, namedForeignKeys={"col1": {"name": "TestFK", "reference": "TableXY(abc)"}}), "CREATE TABLE IF NOT EXISTS TestH (col1 varchar(10), CONSTRAINT TestFK FOREIGN KEY (col1) REFERENCES TableXY(abc));")

    def test_A_I_gen_create_table(self):
        with self.assertRaises(exceptions.InvalidTableNameException):
            standard.gen_create_table("", {"id": "int"})

        with self.assertRaises(exceptions.InvalidTableNameException):
            standard.gen_create_table("  ", {"id": "int"})

        with self.assertRaises(exceptions.DictionaryEmptyException):
            standard.gen_create_table("TestI", {})

        with self.assertRaises(exceptions.InvalidTableNameException):
            standard.gen_create_table("Tes t H", {"col1": "int"})

        with self.assertRaises(exceptions.InvalidTableNameException):
            standard.gen_create_table(" T  es tH", {"col1": "int"})

        with self.assertRaises(exceptions.InvalidTableNameException):
            standard.gen_create_table("Te stH", {"col1": "int", "xyz": "char"})

        with self.assertRaises(exceptions.InvalidTableNameException):
            standard.gen_create_table("Te stH  ", {"col1": "int", "xyz": "char"})

        with self.assertRaises(exceptions.InvalidPrimaryKeyColumn):
            standard.gen_create_table("Test", {'col1': 'int'}, primaryKeys=['col2'])

        with self.assertRaises(exceptions.InvalidForeignKeyColumn):
            standard.gen_create_table("Test", {'col1': 'int'}, foreignKeys={'col2': 'Test(xyz)'})

        with self.assertRaises(exceptions.InvalidUniqueColumn):
            standard.gen_create_table("Test", {'col1': 'int'}, uniqueColumns=['col2'])

        with self.assertRaises(exceptions.InvalidForeignKeyColumn):
            standard.gen_create_table("Test", {'col1': 'int'}, namedForeignKeys={'col2': {'name':"testFK", "reference": "TableABC(def)"}})

        with self.assertRaises(exceptions.InvalidNamedForeignKeyDictionary):
            standard.gen_create_table("Test", {"col1": "int"}, namedForeignKeys={"col1": 42})

        with self.assertRaises(exceptions.InvalidNamedForeignKeyDictionary):
            standard.gen_create_table("Test", {"col1": "int"}, namedForeignKeys={"col1": {}})

        with self.assertRaises(exceptions.InvalidNamedForeignKeyDictionary):
            standard.gen_create_table("Test", {"col1": "int"}, namedForeignKeys={"col1": {'name': 'testFK'}})

        with self.assertRaises(exceptions.InvalidNamedForeignKeyDictionary):
            standard.gen_create_table("Test", {"col1": "int"}, namedForeignKeys={"col1": {'reference': 'anotherTable(aColumn)'}})

    def test_A_J_INVALID_CHARS(self):
        self.assertEqual(standard.INVALID_CHARS, ['!', '"', '#', r'\$', '%', '&', "'", r'\(', r'\)', r'\*', r'\+', ',', '-', '/', ':', ';', '<', '=', '>', r'\?', '@', r'\[', r'\\', r'\]', r'\^', '_', '`', r'\{', r'\|', r'\}', '~', ' ', '\n', '\t'])

    def test_A_K_check_validName(self):
        chars = string.ascii_letters + string.digits

        for i in range(10):
            for char in standard.INVALID_CHARS:
                text = [random.choice(chars) for _ in range(random.randint(10,100))]

                text[random.randint(0, len(text)-1)] = char

                text = ''.join(text)

                self.assertFalse(standard.check_validName(text))

    def test_A_L_check_validName(self):
        chars = string.ascii_letters + string.digits

        for _ in range(50):
            text = ''.join([random.choice(chars) for _ in range(70)])

            self.assertTrue(standard.check_validName(text))

    def test_A_M_gen_update(self):
        self.assertEqual(standard.gen_update("TableA", {"col1": "val1"}), ("UPDATE TableA SET col1=?;", ['val1']))
        self.assertEqual(standard.gen_update("TableA", {"col1": True}), ("UPDATE TableA SET col1=?;", [True]))
        self.assertEqual(standard.gen_update("TableA", {"col1": 42}), ("UPDATE TableA SET col1=?;", [42]))

        self.assertEqual(standard.gen_update("TableA", {"col1": "val1"}, {'col5': 'hello'}), ("UPDATE TableA SET col1=? WHERE col5=?;", ['val1', 'hello']))
        self.assertEqual(standard.gen_update("TableA", {"col1": "val1"}, {'col5': True}), ("UPDATE TableA SET col1=? WHERE col5=?;", ['val1', True]))
        self.assertEqual(standard.gen_update("TableA", {"col1": "val1"}, {'col5': 42}), ("UPDATE TableA SET col1=? WHERE col5=?;", ['val1', 42]))

        self.assertEqual(standard.gen_update("TableA", {"col1": True}, {'col5': 'hello'}), ("UPDATE TableA SET col1=? WHERE col5=?;", [True, 'hello']))
        self.assertEqual(standard.gen_update("TableA", {"col1": True}, {'col5': True}), ("UPDATE TableA SET col1=? WHERE col5=?;", [True, True]))
        self.assertEqual(standard.gen_update("TableA", {"col1": True}, {'col5': 42}), ("UPDATE TableA SET col1=? WHERE col5=?;", [True, 42]))

        self.assertEqual(standard.gen_update("TableA", {"col1": 42}, {'col5': 'hello'}), ("UPDATE TableA SET col1=? WHERE col5=?;", [42, 'hello']))
        self.assertEqual(standard.gen_update("TableA", {"col1": 42}, {'col5': True}), ("UPDATE TableA SET col1=? WHERE col5=?;", [42, True]))
        self.assertEqual(standard.gen_update("TableA", {"col1": 42}, {'col5': 42}), ("UPDATE TableA SET col1=? WHERE col5=?;", [42, 42]))

        self.assertEqual(standard.gen_update("TableA", {"col1": 42}, {'col5': 42, "colB": "welt"}, conditionCombining="OR"), ("UPDATE TableA SET col1=? WHERE col5=? OR colB=?;", [42, 42, 'welt']))
        self.assertEqual(standard.gen_update("TableA", {"col1": 42}, {'col5': 42, "colB": "welt"}), ("UPDATE TableA SET col1=? WHERE col5=? AND colB=?;", [42, 42, 'welt']))

    def test_A_N_gen_update(self):
        with self.assertRaises(exceptions.InvalidTableNameException):
            standard.gen_update("T abl e", {"col1": "val1"})

        with self.assertRaises(exceptions.InvalidTableNameException):
            standard.gen_update("", {"col1": "val1"})

        with self.assertRaises(exceptions.DictionaryEmptyException):
            standard.gen_update("TableXY", {})

    def test_A_O_gen_drop_db(self):
        self.assertEqual(standard.gen_drop_db("DbA"), 'DROP DATABASE DbA;')

    def test_A_P_gen_drop_db(self):
        with self.assertRaises(exceptions.InvalidDatabaseNameException):
            standard.gen_drop_db("")

    def test_A_Q_gen_drop_table(self):
        self.assertEqual(standard.gen_drop_table("TableA"), 'DROP TABLE TableA;')

    def test_A_R_gen_drop_table(self):
        with self.assertRaises(exceptions.InvalidTableNameException):
            standard.gen_drop_table("")

    def test_A_S_gen_select(self):
        self.assertEqual(standard.gen_select("TableS"), ('SELECT * FROM TableS;', []))
        self.assertEqual(standard.gen_select("TableS", ["col1", "xyz"]), ('SELECT col1, xyz FROM TableS;', []))
        self.assertEqual(standard.gen_select("TableS", ["xyz"]), ('SELECT xyz FROM TableS;', []))
        self.assertEqual(standard.gen_select("TableS", conditions={"col5":"hello"}), ('SELECT * FROM TableS WHERE col5=?;', ['hello']))
        self.assertEqual(standard.gen_select("TableS", ["abc", "def"], conditions={"col5":"hello"}), ('SELECT abc, def FROM TableS WHERE col5=?;', ['hello']))
        self.assertEqual(standard.gen_select("TableS", conditions={"col5":"hello", "xyz": "abc"}), ('SELECT * FROM TableS WHERE col5=? AND xyz=?;', ['hello', 'abc']))
        self.assertEqual(standard.gen_select("TableS", ["abc", "def"], conditions={"col5":"hello", "xyz": "abc"}), ('SELECT abc, def FROM TableS WHERE col5=? AND xyz=?;', ['hello', 'abc']))

        self.assertEqual(standard.gen_select("TableS", ["abc", "def"], conditions={"col5":"hello"}, distinct=True), ('SELECT DISTINCT abc, def FROM TableS WHERE col5=?;', ['hello']))

    def test_A_T_gen_select(self):
        with self.assertRaises(exceptions.InvalidTableNameException):
            standard.gen_select("")

    def test_A_U_gen_count_avg_sum(self):
        with self.assertRaises(exceptions.InvalidType):
            standard.gen_count_avg_sum(None, "TableA", "ColumnB")

        with self.assertRaises(exceptions.InvalidTableNameException):
            standard.gen_count_avg_sum(standard.Count_avg_sum_modes.COUNT, "", "ColumnB")

        with self.assertRaises(exceptions.InvalidType):
            standard.gen_count_avg_sum(standard.Count_avg_sum_modes.COUNT, "TableA", None)

    def test_A_V_gen_count(self):
        self.assertEqual(standard.gen_count("TableA"), ("SELECT COUNT(*) FROM TableA;", []))
        self.assertEqual(standard.gen_count("TableA", "myColumn"), ("SELECT COUNT(myColumn) FROM TableA;", []))
        self.assertEqual(standard.gen_count("TableA", "myColumn", {"colB": 123}), ("SELECT COUNT(myColumn) FROM TableA WHERE colB=?;", [123]))

        self.assertEqual(standard.gen_count("TableA", "myColumn", {"colB": 123}, distinct=True), ("SELECT COUNT(DISTINCT myColumn) FROM TableA WHERE colB=?;", [123]))

    def test_A_W_gen_avg(self):
        self.assertEqual(standard.gen_avg("TableA"), ("SELECT AVG(*) FROM TableA;", []))
        self.assertEqual(standard.gen_avg("TableA", "myColumn"), ("SELECT AVG(myColumn) FROM TableA;", []))
        self.assertEqual(standard.gen_avg("TableA", "myColumn", {"colB": 123}), ("SELECT AVG(myColumn) FROM TableA WHERE colB=?;", [123]))

        self.assertEqual(standard.gen_avg("TableA", "myColumn", {"colB": 123}, distinct=True), ("SELECT AVG(DISTINCT myColumn) FROM TableA WHERE colB=?;", [123]))

    def test_A_X_gen_sum(self):
        self.assertEqual(standard.gen_sum("TableA"), ("SELECT SUM(*) FROM TableA;", []))
        self.assertEqual(standard.gen_sum("TableA", "myColumn"), ("SELECT SUM(myColumn) FROM TableA;", []))
        self.assertEqual(standard.gen_sum("TableA", "myColumn", {"colB": 123}), ("SELECT SUM(myColumn) FROM TableA WHERE colB=?;", [123]))

        self.assertEqual(standard.gen_sum("TableA", "myColumn", {"colB": 123}, distinct=True), ("SELECT SUM(DISTINCT myColumn) FROM TableA WHERE colB=?;", [123]))

    def test_A_Y_a_gen_create_stored_procedure(self):
        self.assertEqual(standard.gen_create_stored_procedure("Test", "SELECT * FROM TableA;"), "CREATE PROCEDURE Test\nAS\nSELECT * FROM TableA;\nGO;")
        self.assertEqual(standard.gen_create_stored_procedure("Test", "SELECT * FROM TableA;", {'param': 'varchar(10)'}), "CREATE PROCEDURE Test @param varchar(10)\nAS\nSELECT * FROM TableA;\nGO;")
        self.assertEqual(standard.gen_create_stored_procedure("Test", "SELECT * FROM TableA;", {'param': 'varchar(10)', 'param2': 'varchar(20)'}), "CREATE PROCEDURE Test @param varchar(10), @param2 varchar(20)\nAS\nSELECT * FROM TableA;\nGO;")

    def test_A_Y_b_gen_create_stored_procedure(self):
        with self.assertRaises(exceptions.InvalidType):
            standard.gen_create_stored_procedure(123, "SELECT * FROM TableA;")

        with self.assertRaises(exceptions.InvalidType):
            standard.gen_create_stored_procedure("Test", 456)

        with self.assertRaises(exceptions.InvalidType):
            standard.gen_create_stored_procedure("Test", "SELECT * FROM TableA;", ['abc','def'])

        with self.assertRaises(exceptions.InvalidType):
            standard.gen_create_stored_procedure("Test", "SELECT * FROM TableA;", {123:'456'})

        with self.assertRaises(exceptions.InvalidType):
            standard.gen_create_stored_procedure("Test", "SELECT * FROM TableA;", {'123':456})

    def test_A_Z_a_gen_exec_procedure(self):
        self.assertEqual(standard.gen_exec_procedure("Test"), ("EXEC Test;", []))
        self.assertEqual(standard.gen_exec_procedure("Test", {'param': 'val1'}), ("EXEC Test @param=?;", ['val1']))
        self.assertEqual(standard.gen_exec_procedure("Test", {'param': 'val1', 'param2': 'val2'}), ("EXEC Test @param=?, @param2=?;", ['val1', 'val2']))

    def test_A_Z_b_gen_exec_procedure(self):
        with self.assertRaises(exceptions.InvalidType):
            standard.gen_exec_procedure(123)

        with self.assertRaises(exceptions.InvalidType):
            standard.gen_exec_procedure("Test", ['abc','def'])

    def test_B_A_b_gen_drop_procedure(self):
        self.assertEqual(standard.gen_drop_procedure("Test"), "DROP PROCEDURE Test;")

    def test_B_A_b_gen_drop_procedure(self):
        with self.assertRaises(exceptions.InvalidType):
            standard.gen_drop_procedure(123)



from sqlIntuitive.exceptions.parameterExceptions.subexceptions.invalidConstraintExceptions.baseException import InvalidConstraint

class InvalidPrimaryKeyColumn(InvalidConstraint):
    def __init__(self, columnName):
        super().__init__()

        self.columnName = columnName

    def __str__(self):
        return f"{self.columnName} not in columns."

class InvalidForeignKeyColumn(InvalidConstraint):
    def __init__(self, columnName):
        super().__init__()

        self.columnName = columnName

    def __str__(self):
        return f"{self.columnName} not in columns."

class InvalidUniqueColumn(InvalidConstraint):
    def __init__(self, columnName):
        super().__init__()

        self.columnName = columnName

    def __str__(self):
        return f"{self.columnName} not in columns."

class InvalidNamedForeignKeyDictionary(InvalidConstraint):
    def __init__(self, columnName):
        super().__init__()

        self.columnName = columnName

    def __str__(self):
        return f"Values for column {self.columnName} are not defined."
